#! /usr/bin/env python
import os
import argparse
import pickle
import logging

import numpy as np
import pandas as pd
from tqdm.auto import tqdm

import pymc3 as pm

from sep241util import logger, setup_logging
from sep241util import read_job_data, read_results, write_bed
from sep241util import MissingData
from sep241util import posterior_mode_weights
from sep241util import check_length_distribution_flip
from sep241model import get_length_dist_modes


desc = """Export deconvolution results for each individual event in the data.
Each end of a fragment is considered an event. The output format is one
bed-like tab-separated text table per channel: ``c1.bed`` and ``c2.bed``.
The columns of the two tables correspond to:

    chrom, chromStart, chromEnd, name, likelihood, location-marginal likelihood

Where *likelihood* corresponds to the likelihood of the event to originate
from the respective target and *location marginal likelihood* ignores the
length of the associated fragment and only uses the genomic-location-specific
target likelihood. The one-dimensional
genomic-location-specific target likelihood can also be exported with
sep241mkbw.
"""
parser = argparse.ArgumentParser(description=desc)
parser.add_argument(
    "jobdata",
    metavar="jobdata-file",
    type=str,
    nargs="?",
    help="Jobdata with cuts per interval and workchunk ids.",
)
parser.add_argument(
    "-l",
    "--log",
    dest="logLevel",
    choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    default="INFO",
    help="Set the logging level (default=INFO).",
    metavar="LEVEL",
)
parser.add_argument(
    "--logfile",
    help="Write detailed log to this file.",
    type=str,
    metavar="logfile",
)
parser.add_argument(
    "-o",
    "--out",
    help="Output directory (default is the path of the jobdata).",
    type=str,
    metavar="out_dir",
)
parser.add_argument(
    "--no-check",
    help="Do not test for flipped length distributions.",
    action="store_true",
)
parser.add_argument(
    "--no-progress", 
    help="Do not show progress.", 
    action="store_true",
)
parser.add_argument(
    "--force",
    help="Make bigwigs even if some results are missing.",
    action="store_true",
)
parser.add_argument(
    "--exclude-flipped",
    help="Exclude results from workchunks with flipped length distribution.",
    action="store_true",
)


def likelihoods_per_cut(workdata, map_results, progress=True):
    df_list = list()
    last_name = ""

    max_log_value = None
    missing_data_wg = set()

    for name, dat in tqdm(
        workdata.iterrows(), total=len(workdata), desc="intervals", disable=not progress
    ):
        wg = dat["workchunk"]
        maxlle = map_results.get(dat["workchunk"], None)
        if maxlle is None:
            if wg not in missing_data_wg:
                logger.warning("Data of workchunk %d is None.", wg)
                missing_data_wg.add(wg)
            continue
        if f"f_c1_{name}" not in maxlle.keys():
            logger.warning("Marginal likelihood results not found in workchunk %d.", wg)
            continue
        if max_log_value is None:
            max_log_value = np.log(np.finfo(maxlle[f"f_c1_{name}"].dtype).max)
        start_idx = dat["stich_start"][0]
        end_idx = dat["stich_end"][0]
        seq = dat["seqname"]

        df = dat["cuts"].iloc[start_idx:end_idx, :].copy()
        locations = df["location"].values.astype(int)

        idx = dat["cuts"]["location"].rank(method="dense").astype(int) - 1

        df["log_location_marginal_c1"] = maxlle[f"f_c1_{name}"][idx][start_idx:end_idx]
        df["log_location_marginal_c2"] = maxlle[f"f_c2_{name}"][idx][start_idx:end_idx]

        df["workchunk"] = wg
        df["name"] = name
        df_list.append(df)

    return pd.concat(df_list, axis=0)


def main():
    args = parser.parse_args()
    logging.getLogger("filelock").setLevel(logging.ERROR) # aesara produces a lot
    setup_logging(args.logLevel, args.logfile)
    logger.debug("Loglevel is on DEBUG.")

    logger.info("Reading jobdata.")
    workdata = read_job_data(args.jobdata)

    logger.info("Reading deconvolution results.")
    try:
        map_results = read_results(
            args.jobdata, workdata, progress=not args.no_progress, error=not args.force
        )
    except MissingData:
        raise MissingData(
            "Results are missing. Complete missing work chunks or pass --force to "
            "call peaks regardless."
        )
    if not args.no_check:
        bad_wgs = check_length_distribution_flip(workdata, map_results)
        if args.exclude_flipped:
            logger.info("Removing flipped workgroups from results.")
            for idx in bad_wgs:
                del map_results[idx]

    logger.info("Getting location marginal likelihoods.")
    events = likelihoods_per_cut(workdata, map_results, progress=not args.no_progress)

    logger.info("Producing posterior length marginal likelihood distributions.")
    dc_args = next(iter(map_results.values()))["arguments"]
    length_comps = get_length_dist_modes(
        dc_args.length_dist_modes, dc_args.length_dist_mode_sds
    )()
    w_c1_posterior, w_c2_posterior = posterior_mode_weights(workdata, map_results)

    logger.info("Computing length marginal likelihoods for component 1.")
    mdist_c1 = pm.Mixture.dist(
        w=w_c1_posterior / np.sum(w_c1_posterior), comp_dists=length_comps
    )
    events["log_length_marginal_c1"] = mdist_c1.logp(events["length"]).eval()

    logger.info("Computing length marginal likelihoods for component 2.")
    mdist_c2 = pm.Mixture.dist(
        w=w_c2_posterior / np.sum(w_c2_posterior), comp_dists=length_comps
    )
    events["log_length_marginal_c2"] = mdist_c2.logp(events["length"]).eval()

    logger.info("Postprocessing.")
    events["likelihood_c1"] = np.exp(
        events["log_length_marginal_c1"] + events["log_location_marginal_c1"]
    )
    events["likelihood_c2"] = np.exp(
        events["log_length_marginal_c2"] + events["log_location_marginal_c2"]
    )
    events["marginal_c1"] = np.exp(events["log_location_marginal_c1"])
    events["marginal_c2"] = np.exp(events["log_location_marginal_c2"])
    events["bed_end"] = events["location"] + 1

    if args.out is None:
        out_path, old_file = os.path.split(args.jobdata)
    else:
        out_path = args.out
        try:
            os.mkdir(out_path)
        except FileExistsError:
            pass

    out_base = os.path.join(out_path, "events_deconv_")

    out_c1 = out_base + "c1.bed"
    columns = ["seqname", "location", "bed_end", "barcode"]
    if "barcode" not in events.columns:
        logger.info("Generating event names.")
        n = len(events)
        nchars = int(np.ceil(np.log10(n)))
        event_numbers = pd.Series(range(n)).astype(str).str.zfill(nchars)
        events["barcode"] = "event_" + event_numbers.values
    logger.info("Writing results to %s", out_c1)
    write_bed(
        events[columns + ["likelihood_c1", "marginal_c1"]], out_c1,
    )
    out_c2 = out_base + "c2.bed"
    logger.info("Writing results to %s", out_c2)
    write_bed(
        events[columns + ["likelihood_c2", "marginal_c2"]], out_c2,
    )
    logger.info("Finished sucesfully.")


if __name__ == "__main__":
    main()



from scipy.sparse import csc_matrix
import numpy as np
import theano
import theano.tensor as at
from pymc3.gp.util import infer_shape
import warnings

THRESHOLD_DEFAULT = 1e-8

class SparseCov:
    R"""
    A callable wrapper class for covariance functions that represents the covariance matrix
    sparsely under certain assumptions:
    
    The covariance function k(x, y) must have a positive range, only dependent on the 
    distance d = abs(x - y), and be strictly decreasing with respect to d. 
    Further, the data must be strictly increasing, 1-dimensional, and contain integers
    (divisible by 1, not necessarily type int). Since the data is strictly increasing, for
    any data point x, as y becomes farther away in terms of index, d must be increasing.
    Since k(x, y) is strictly decreasing with respect to d, for any data point x, once we
    find a y for which k(x, y) <= threshold, every data point beyond y must also satisfy
    k(x, y) <= threshold. 
    
    So if we choose a threshold at which we consider the covariance
    negligible, the significant data will be clustered around the diagonal. So we can compute
    the covariance matrix in O(nm) time and memory, where n is the length of the data and m
    is the average number of points y satisfying k(x, y) <= threshold across all x.

    :param cov_func: A pymc3 covariance function. Must have a positive range and be strictly
        decreasing with respect to the distance between points.
    :type cov_func: pymc3.gp.cov.Covariance
    :param threshold: Do not calculate or represent any covariances k <= threshold.
        Defaults to 1e-8.
    :type threshold: float, optional
    """

    def __init__(self, cov_func, threshold=THRESHOLD_DEFAULT):
        if threshold < 0:
            raise ValueError('Threshold must be positive.')

        a = at.dvector()
        a.tag.test_value = np.ones(2)

        # We can think of the covariance function k(x, y) as a monotonically decreasing function of the
        # distance between x and y, notated k(d). Here the covariance function is structured so
        # it transforms a vector of distances.
        with warnings.catch_warnings():
            # pymc covariance functions interpret symbolic matrices as having the wrong input
            # dimensions, throwing an unnecesssary warning.
            warnings.simplefilter("ignore")
            self.cov_func = theano.function([a], cov_func(np.array([[0.0]]), a.reshape((-1, 1))).flatten())


        self.threshold = threshold
        
        # Since k is monotonically decreasing, k(d) <= threshold implies k(d + c) <= threshold for any
        # positive constant c. We assume integer data, so we only really need the first integer d for
        # which k(d) <= threshold.
        self.inv = np.ceil(self._find_inverse())

    def _find_inverse(self, upper=1e+15):
        R"""
        Finds the minimum distance d such that the k(d) <= threshold for all distances equal or greater.
        Searches between 0 and upper.
        Parameters
        ----------
        upper: float
            Assume that the inverse is less than this value. Must be greater than 0. Defaults to 1e+15.
        """

        lower = 0.0

        # Invariant:
        # k(lower) > threshold
        # k(upper) <= threshold
        # lower < upper
        while upper > np.nextafter(lower, np.inf):
            # If we are in the loop, there is at least one representable floating point
            # between upper and lower.

            middle = (upper / 2.0) + (lower / 2.0)
            
            # In case of numerical issues, make sure lower < middle < upper. Because there is a
            # floating point between upper and lower, we can make this inequality exclusive.
            if middle <= lower:
                middle = np.nextafter(lower, np.inf)
            elif middle >= upper:
                middle = np.nextafter(upper, -np.inf)

            sim = self.cov_func(np.array([middle]))
            if sim > self.threshold:
                # k(middle) > threshold.
                lower = middle
            else:
                # k(middle) <= threshold
                upper = middle
            # Since we had lower < middle < upper, and then we set one of lower or upper to middle,
            # when the loop repeats or terminates we have lower < upper.

        # When the loop terminates, we have:
        # k(lower) > threshold
        # k(upper) <= threshold
        # lower < upper
        # upper <= np.nextafter(lower, np.inf)

        # Simplifying the last two statements, lower < upper <= np.nextafter(lower, np.inf), so
        # upper = np.nextafter(lower, np.inf). So upper is the first distance d for which
        # k(d) <= threshold and lower is the last distance d for which k(d) > threshold.
        return upper

    def _find_spine(self, X):
        R"""
        For each datapoint in X, finds the number of the number of datapoints with distance d such that
        k(d) > threshold. Returns an array for the number of datapoints to the left within this distance
        and an array for the number of datapoints to the right within this distance.
        Parameters
        ----------
        X: array-like
            Function input values.
        """
        n = infer_shape(X)
        
        # Side left to not include the exact threshold.
        rightspine = np.searchsorted(X, X + self.inv, side='left') - np.arange(1, n + 1)
        reverse = -np.flip(X)
        leftspine = np.flip(np.searchsorted(reverse, reverse + self.inv, side='left') - np.arange(1, n + 1))
        return leftspine, rightspine

    def _align(self, X):
        R"""
        Finds all the pairs of entries in X with distance d such that k(d) > threshold. Returns a tuple containing
        the data, indices and indptr that place the distance between x[i] and x[j] in locations (i, j) and (j, i)
        in a csc matrix.
        Parameters
        ----------
        X: array-like
            Function input values.
        """
        X = X.flatten().astype(np.int64)
        n = infer_shape(X)

        # Number of dense entries left of the diagonal for each row, number of entries right of the diagonal for each row
        nnzleft, nnzright = self._find_spine(X)

        # Total number of dense entries for each row
        nnzrow = nnzleft + nnzright + 1

        # Number of entries preceding each row
        indptr = np.append(0, np.cumsum(nnzrow))

        # Imagine assigning indices one a at time. When assigning an index to an entry, the next entry in the same row
        # is one index greater.
        differences = np.ones(np.sum(nnzrow), dtype=np.int64)

        # When assigning the index of the first entry of a row, move left one square for each entry right of the
        # diagonal in the previous row and one square for each entry left of the diagonal in this row. Still move
        # right one index because the diagonal shifts one index. Also subtract 1 from the first entry of differences
        # to start at index 0.
        differences[indptr[:-1]] -= np.append(1, nnzleft[1:] + nnzright[:-1])
        indices = np.cumsum(differences, out=differences)

        # A_ij = |X[i] - X[j]|
        data = np.abs(np.repeat(X, nnzrow) - X[indices])
        return data, indices, indptr

    def auto_jitter(self, matrix):
        R"""
        Calculates the jitter necessary to guarantee positive definiteness of the
        sparsified covariance matrix. Assumes that the matrix would be positive
        semi-definite before the sparsification.

        :param matrix: The NxN sparsified covariance matrix.
        :type matrix: scipy.sparse.spmatrix
        :return: jitter to guarantee positive definiteness.
        :rtype: float
        """
        # A Hermitian strictly diagonally dominant matrix with non-negative diagonals
        # is PD. Source: https://en.wikipedia.org/wiki/Diagonally_dominant_matrix.
        # Consider the sparsification matrix. It has zeros wherever we leave
        # the original matrix alone and negative values where we set the original
        # matrix to zero, which can only be in off diagonal entries. If we set the
        # diagonal of the sparsification matrix greater than the magnitude of the
        # sum of corresponding rows, the sparsification matrix is real, strictly 
        # diagonally dominant, and has non-negative diagonals, and is therefore PD.
        # Since the original covariance matrix is PSD and the sparsification matrix
        # is PD, their sum is PD.

        n = infer_shape(matrix)

        # rows and cols order the indices of the non-zero entries of the matrix
        # left to right of row 0, left to right of row 1, and so on.
        rows, cols = matrix.nonzero()

        # We want to find the number of zeros to the left of the leftmost non-zero entry and
        # to the right of the rightmost entry in each row. We can assume that every row has
        # at least one element because the diagonal is assumed to be non-zero. The ith index
        # of np.diff(rows) == 1 is True iff the the row changes between the ith entry and the
        # (i+1)th entry. So np.diff(rows) == 1 is a mask for the entries before each row change,
        # i.e. the last nonzero entry of each row except the last row, since there is no row
        # change from the last row to a nonexistent row.

        # The first nonzero entry of each row occurs one entry after the last nonzero entry
        # of the previous row. Since np.diff(rows) == 1 is a mask for the last entries of
        # each row except the last, by padding the mask with a True in the beginning, we
        # offset the mask by one, so the new mask accesses the first entries of each row,
        # including the first non-zero entry of row 0, which must be in row 0 column 0. The
        # column indices of the first non-zero entries of each row is the same as the number
        # of zero entries on the left side. 
        leftspine = cols[np.append(True, np.diff(rows) == 1)]

        # By padding the mask with a True at the end, the new mask accesses the last nonzero
        # entries of each row, including the last non-zero entry of the last row, which must
        # be in row n - 1 column n - 1. The column indices of the last non-zero entries of
        # each row give the number of zero or non-zero entries to the left of the last non-zero
        # entries of each row, not including the entry itself, so if i is the column index
        # of the last non-zero entry for a row, the number of zero entries to the right of the
        # last non-zero entry is n - i - 1.
        rightspine = n - (cols[np.append(np.diff(rows) == 1, True)] + 1)

        # The inverse is defined as the minimum distance that the covariance function
        # transforms to a value less than or equal to the threshold. Distances that get
        # thrown out must be at least the inverse. Our strategy is to assume the worst:
        # that the distances increase as little as possible. The largest distance that we
        # may have to assume is for the case that the first or last row has n - 1 zeros.
        # In this case the largest distance is self.inv + n - 2.
        distances = np.arange(self.inv, self.inv + n - 1)

        # The associated covariances for all the possible distances.
        costs = self.cov_func(distances)
        costs = np.append(0, costs)

        # A series of m negatives cannot have a greater cost than
        # k(inv) + ... + k(inv + m - 1). So
        # seriescost[m] is the maximum cost of a series of m negatives.
        seriescosts = np.cumsum(costs)

        # Rowcosts gives the maximum amount subtracted in each row by the sparsification.
        rowcosts = seriescosts[leftspine] + seriescosts[rightspine]

        # For consistent variance, choose the same jitter for the whole diagonal.
        return np.max(rowcosts) + 1e-6

    def __call__(self, X):
        R"""
        Returns a sparse covariance matrix, only calculating entries greater than the threshold.
        Because the input data must be strictly increasing, all dense entries of the resulting
        covariance matrix will be located around the main diagonal.
        Parameters
        ----------
        X: array-like
            Function input values.
        """
        # Get the distance of all data points to compare
        data, indices, indptr = self._align(X)

        # Transform the distances with the covariance function
        data = self.cov_func(data)

        return csc_matrix((data, indices, indptr))



#! /usr/bin/env python
import os
import argparse
import numpy as np
import pandas as pd
import scipy

from threadpoolctl import threadpool_limits
from tqdm.auto import tqdm
from sep241util import logger, setup_logging, detect_cores
import sep241prep as spr
import sep241peakcalling as sep241peakcalling


desc = "Use 2for1seperator peak calling based on KDE of two single antibody CUT&Tag files."
parser = argparse.ArgumentParser(description=desc)
parser.add_argument(
    "c1_bed_file",
    metavar="reads_of_c1.bed",
    type=str,
    help="Bed file with single antibody reads of component 1.",
)
parser.add_argument(
    "c2_bed_file",
    metavar="reads_of_c2.bed",
    type=str,
    help="Bed file with single antibody reads of component 2.",
)
parser.add_argument(
    "out", help="Output directory.", type=str, metavar="out_dir",
)
parser.add_argument(
    "-l",
    "--log",
    dest="logLevel",
    choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    default="INFO",
    help="Set the logging level (default=INFO).",
    metavar="LEVEL",
)
parser.add_argument(
    "--logfile",
    help="Write detailed log to this file.",
    type=str,
    metavar="logfile",
)
parser.add_argument(
    "--blacklisted-seqs",
    help="Sequences to exclude from the deconvolution (default=chrM).",
    type=str,
    default=["chrM"],
    nargs="+",
    metavar="chrN",
)
parser.add_argument(
    "--kde-bw",
    help="Bandwidth (sigma) for kernel density estimate (KDE) used for interval selection (default=200).",
    type=float,
    default=200,
    metavar="float",
)
parser.add_argument(
    "--c1-min-peak-size",
    help="Minimal number of bases per peak for component 1 (default=100).",
    type=int,
    default=100,
    metavar="int",
)
parser.add_argument(
    "--c2-min-peak-size",
    help="Minimal number of bases per peak for component 2 (default=400).",
    type=int,
    default=400,
    metavar="int",
)
parser.add_argument(
    "--c1-smooth",
    help="Apply gaussian filter with this standard deviation to component 1 (default=0).",
    type=float,
    default=0,
    metavar="float",
)
parser.add_argument(
    "--c2-smooth",
    help="Apply gaussian filter with this standard deviation to component 2 (default=2000).",
    type=float,
    default=2000,
    metavar="float",
)
parser.add_argument(
    "--fraction-in-peaks",
    help="Fraction of reads that are expected to be in peaks (default=0.5).",
    type=float,
    default=0.5,
    metavar="float",
)
parser.add_argument(
    "--fraction-overlap",
    help="""If more than this fraction of a peak overlaps with a peak of the
    other target than it is considered an overlapping region (default=0.5).""",
    type=float,
    default=0.5,
    metavar="float",
)
parser.add_argument(
    "--uncorrected",
    help="Do not correct cut ratio estimate with Bayesian prior.",
    action="store_true",
)
parser.add_argument(
    "--span",
    help="Resolution in number of base pairs (default=10).",
    type=int,
    default=10,
    metavar="int",
)
parser.add_argument(
    "--no-progress", help="Do not show progress.", action="store_true",
)
parser.add_argument(
    "--cores", help="Number of CPUs to use.", type=int, default=0, metavar="int",
)


def read_bed(file_path):
    header = {0: "seqname", 1: "start", 2: "end"}
    bed_content = pd.read_csv(file_path, delimiter="\t", header=None).iloc[:, :3]
    bed_content.columns = ["seqname", "start", "end"]
    return bed_content


def write_bed(bed_df, out_path):
    bed_df.to_csv(out_path, sep="\t", header=False, index=False)


def envents_from_intervals(interval_df):
    id_vars = set(interval_df.columns) - {"start", "end"}
    df = interval_df.melt(id_vars=id_vars, value_name="location")
    return df


def name_peaks(df, prefix):
    name_dat = pd.DataFrame(
        list(df.index.str.split("_")),
        columns=["seqname", "interval", "is_peak"],
        index=df.index,
    )
    df["name"] = (
        str(prefix) + "_"
        + name_dat["seqname"] + "_"
        + name_dat["interval"]
    )
    return df


def events_dict_from_file(c1_file, c2_file):
    df_c1 = read_bed(c1_file)
    df_c2 = read_bed(c2_file)

    events_c1 = envents_from_intervals(df_c1)
    events_c2 = envents_from_intervals(df_c2)

    ebs_c1 = {
        seqname: seq_events for seqname, seq_events in events_c1.groupby("seqname")
    }
    ebs_c2 = {
        seqname: seq_events for seqname, seq_events in events_c2.groupby("seqname")
    }

    return ebs_c1, ebs_c2


def deconv_like_kde(
    ebs_c1,
    ebs_c2,
    c1_sigma=None,
    c2_sigma=2000,
    step=10,
    kde_bw=200,
    blacklisted=list(),
):

    sequences = set(ebs_c1.keys()).union(ebs_c2.keys()) - set(blacklisted)
    result_dfs = list()
    ntotal = len(sequences)
    signal_c1_list_global = list()
    signal_c2_list_global = list()

    for i, seqname in enumerate(sequences):
        logger.info(f"Making KDE of {seqname} [{i+1}/{ntotal}].")

        events_c1 = ebs_c1.get(seqname, pd.DataFrame())
        events_c2 = ebs_c2.get(seqname, pd.DataFrame())

        events_c1["true_origin"] = "c1"
        events_c2["true_origin"] = "c2"

        cuts_c1 = events_c1["location"].values
        cuts_c2 = events_c2["location"].values

        all_cuts = np.concatenate([cuts_c1, cuts_c2])
        grid = spr.full_kde_grid(all_cuts)
        cut_idx = all_cuts - grid.min()

        _, density_c1 = spr.get_kde(cuts_c1, kde_bw=kde_bw, grid=grid)
        density_c1 *= len(events_c1)
        _, density_c2 = spr.get_kde(cuts_c2, kde_bw=kde_bw, grid=grid)
        density_c2 *= len(events_c2)

        comb_df = pd.DataFrame(
            {
                "seqname": seqname,
                "interval": seqname,
                "location": grid[::step],
                "c1": density_c1[::step],
                "c2": density_c2[::step],
            }
        )

        if c1_sigma:
            comb_df["c1 smooth"] = scipy.ndimage.gaussian_filter1d(
                comb_df["c1"], int(c1_sigma / step)
            )
        if c2_sigma:
            comb_df["c2 smooth"] = scipy.ndimage.gaussian_filter1d(
                comb_df["c2"], int(c2_sigma / step)
            )

        signal_c1_list_global.append(density_c1[cut_idx])
        signal_c2_list_global.append(density_c2[cut_idx])
        result_dfs.append(comb_df)
    comb_data = pd.concat(result_dfs, axis=0)

    return comb_data, signal_c1_list_global, signal_c2_list_global


def main():
    args = parser.parse_args()
    setup_logging(args.logLevel, args.logfile)
    logger.debug("Loglevel is on DEBUG.")
    cores = detect_cores(args.cores)
    logger.info("Limiting computation to %i cores.", cores)
    threadpool_limits(limits=int(cores))
    os.environ["NUMEXPR_MAX_THREADS"] = str(cores)

    logger.info("Reading %s and %s", args.c1_bed_file, args.c2_bed_file)
    ebs_c1, ebs_c2 = events_dict_from_file(args.c1_bed_file, args.c2_bed_file)

    comb_data, signal_c1_list_global, signal_c2_list_global = deconv_like_kde(
        ebs_c1,
        ebs_c2,
        c1_sigma=args.c1_smooth,
        c2_sigma=args.c2_smooth,
        step=args.span,
        kde_bw=args.kde_bw,
        blacklisted=args.blacklisted_seqs,
    )

    logger.info("Calling peaks.")
    peaks_c1, peaks_c2, overlaps = sep241peakcalling.call_peaks(
        comb_data,
        signal_c1_list_global,
        signal_c2_list_global,
        work_coverage=1.0,
        fraction_in_peaks=args.fraction_in_peaks,
        fraction_overlap=args.fraction_overlap,
        c1_min_peak_size=args.c1_min_peak_size,
        c2_min_peak_size=args.c2_min_peak_size,
        uncorrected=args.uncorrected,
        span=args.span,
        cores=cores,
        progress=not args.no_progress,
    )

    bed_c1 = sep241peakcalling.inlude_auc(name_peaks(peaks_c1, "c1"))
    bed_c2 = sep241peakcalling.inlude_auc(name_peaks(peaks_c2, "c2"))

    out_path = args.out
    try:
        os.mkdir(out_path)
    except FileExistsError:
        pass

    out_base = os.path.join(out_path, "peaks_")
    logger.info("Writing results to %s...", out_base)
    sep241peakcalling.write_peaks(bed_c1, bed_c2, overlaps, out_path)
    logger.info("Finished sucesfully.")


if __name__ == "__main__":
    main()



import sys
import os
path = os.path.dirname(os.path.abspath(__file__))
sys.path.append(path)

import sep241
import sep241.sep241util
import sep241.sep241covariance
import sep241.sep241latent
import sep241.sep241model
import sep241.sep241prep
import sep241.sep241peakcalling
import sep241.sep241events
import sep241.sep241mkbw
import sep241.sep241mkdt


import numpy as np
import pymc3 as pm
from pymc3.gp.util import infer_shape
import theano.sparse as ts
from sep241covariance import SparseCov
from scipy.sparse.linalg import spsolve, splu
from scipy.linalg import LinAlgError
from scipy.sparse import diags
from sksparse.cholmod import cholesky


class SparseLatent:
    R"""
    This class is like pymc3.gp.Latent, except it uses a sparse covariance matrix to save time
    and memory and only supports prior. cov_func must be an instance of SparseCov.
    See documentation of SparseCov for use conditions.

    :param mean_func: The mean function.
    :type mean_func: pymc3.gp.mean.Mean
    :param cov_func: The covariance function. See SparseCov.
    :type cov_func: SparseCov
    """

    def __init__(self, mean_func, cov_func):
        if not isinstance(cov_func, SparseCov):
            raise ValueError('Covariance function must be wrapped by SparseCov.')
        self.mean_func = mean_func
        self.cov_func = cov_func

    def _format_initval(self, initval, L):
        if initval is None:
            return None
        return spsolve(L, initval, 'NATURAL').reshape(-1, 1)

    def _parametrization(self, X, jitter='auto', initval=None):
        size = infer_shape(X)
        mu = self.mean_func(X)
        cov = self.cov_func(X)

        if jitter == 'auto':
            jitter = self.cov_func.auto_jitter(cov)
        factor = cholesky(cov, beta=jitter, mode='simplicial', ordering_method='natural')
        L = factor.L()
        initval = self._format_initval(initval, L)
        return size, mu, L, initval

    def _build_prior(self, name, X, reparameterize=True, jitter='auto', initval=None):
        if reparameterize:
            size, mu, L, initval = self._parametrization(X, jitter=jitter, initval=initval)
            v = pm.Normal(name + '_rotated_', 0, 1, shape=(size, 1),
                          testval=initval)
            f = pm.Deterministic(name, mu + ts.structured_dot(L, v).flatten())
        else:
            size = infer_shape(X)
            mu = self.mean_func(X)
            cov = self.cov_func(X)
            f = pm.Normal(name, 0, 1, shape=size, testval=initval)
        return f

    def prior(self, name, X, reparameterize=True, jitter='auto', initval=None):
        R"""
        Returns the GP prior distribution evaluated over the input
        locations `X`.
        This is the prior probability over the space
        of functions described by its mean and covariance function.

        :param name: Name of the random variable.
        :type name: string
        :param X: Function input values.
        :type X: array-like
        :param reparameterize: Re-parameterize the distribution by rotating the random
            variable by the Cholesky factor of the covariance matrix.
        :type reparameterize: bool
        :param jitter: A correction added to the diagonal of the covariance matrix for numerical
            stability. If set to 'auto', computes a minimum to ensure the covariance
            matrix is positive semi-definite.
            Defaults to 'auto'.
        :type jitter: scalar or 'auto'
        :param initval: Initial value for rotated distribution. Defaults to None.
        :type initval: array-like, optional
        """
        f = self._build_prior(name, X, reparameterize, jitter, initval=None)
        self.X = X
        self.f = f
        return f



import warnings

import os
import re
import logging
import pickle

import numpy as np
import scipy
import pandas as pd
from tqdm.auto import tqdm

from sklearn.gaussian_process import GaussianProcessRegressor
import sklearn.gaussian_process.kernels as kernels
from sklearn.linear_model import LinearRegression
from scipy.stats import lognorm


class NoData(Exception):
    pass


class MissingData(Exception):
    pass


class BEDFormatError(Exception):
    pass


class NotARegion(Exception):
    pass

class BarcodeError(Exception):
    pass


logger = logging.getLogger("2for1seperator")


def setup_logging(level, logfile=None):
    logger.propagate = False
    logging.basicConfig(
        format="[%(asctime)s] [%(levelname)-8s] %(message)s",
        level=logging.getLevelName(level),
    )
    ch = logging.StreamHandler()
    ch.setLevel(logging.getLevelName(level))
    formatter = logging.Formatter("[%(asctime)s] [%(levelname)-8s] %(message)s")
    ch.setFormatter(formatter)
    if logfile:
        fh = logging.FileHandler(logfile)
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(formatter)
        logger.addHandler(fh)
    logger.addHandler(ch)


def set_flag(flag, value=None):
    """
    Sets or overwrites the theano `flag` in the environment variable 'THEANO_FLAGS'.

    :param flag: The flag name that is to be overwritten or set.
    :param value: The value to be assigned to the flag. If it is
        `None` then `flag` will be pasted as is into 'THEANO_FLAGS'.
    :return: The new value of 'THEANO_FLAGS'.
    """
    if not isinstance(flag, str):
        raise TypeError("The argument `flag` needs to be a string.")
    flagString = os.getenv("THEANO_FLAGS")

    if value is None:
        if flagString:
            flag = flagString + "," + flag
        os.environ["THEANO_FLAGS"] = flag
        return newFlagString

    if not isinstance(value, str):
        raise TypeError("The argument `value` needs to be a string or `None`.")

    if flagString:
        oldFlags = flagString.split(",")
    else:
        oldFlags = []
    flagTag = flag + "="
    newFlags = [s for s in oldFlags if flagTag not in s]
    newFlags.append(flagTag + value)
    newFlagString = ",".join(newFlags)
    os.environ["THEANO_FLAGS"] = newFlagString
    return newFlagString


def format_cuts(locations_ds):
    """Formats cuts as represented in the model."""

    interleave_index = locations_ds.rank(method="dense").astype(int) - 1

    cuts_per_pos = locations_ds.value_counts(sort=False).sort_index()
    midpoints = (cuts_per_pos.index[:-1] + cuts_per_pos.index[1:]) / 2
    rectangle_width = np.diff(
        np.concatenate([[locations_ds.min()], midpoints, [locations_ds.max()]])
    )
    # ends have only half of expected width
    rectangle_width[0] += rectangle_width[0]
    rectangle_width[-1] += rectangle_width[-1]
    log_loc_weight = np.log(rectangle_width)

    location_shift = np.median(cuts_per_pos.index).astype(int)  # for numeric ease
    unique_locations = np.array(cuts_per_pos.index - location_shift)

    return unique_locations, log_loc_weight, interleave_index.values


def write_bed(bed_df, out_path):
    bed_df.to_csv(out_path, sep="\t", header=False, index=False)


def detect_cores(cores=None):
    import multiprocessing
    if not cores:
        cores = multiprocessing.cpu_count()
        logger.info("Detecting %s compute cores.", cores)
    return cores


def read_region_string(
    feature, short_seqname=True, format_string=r"([^:]+):([0-9,_]+)-([0-9,_]+)"
):
    region_format = re.compile(format_string)
    is_region = region_format.match(feature)
    if not is_region:
        raise NotARegion(
            f"The passed feature `{feature}` does not match the region format seqname:start-end."
        )
    seq, start, end = is_region.groups()
    start = int(re.sub("[^0-9]", "", start))
    end = int(re.sub("[^0-9]", "", end))
    if short_seqname and seq.startswith("chr"):
        seq = seq[3:]
    return seq, start, end


def length_dist(
    weights,
    sigmas=[0.4, 0.18, 0.115, 0.085],
    modes=[70, 200, 400, 600],
    x=np.linspace(1, 800, 500),
):
    mus = [np.log(modes[i]) + (s ** 2) for i, s in enumerate(sigmas)]
    dists = [
        weights[i] * lognorm.pdf(x, sigmas[i], scale=np.exp(mus[i]), loc=0)
        for i in range(len(weights))
    ]
    return np.sum(np.stack(dists), axis=0)


def posterior_length_dists(
    map_results, x=np.linspace(1, 800, 500), prior=True, progress=False
):
    df_list = list()
    for workchunk, maxlle in tqdm(
        map_results.items(),
        desc="work chunk",
        total=len(map_results),
        disable=~progress,
    ):
        ndfp = pd.DataFrame(
            {
                "density": length_dist(maxlle["weight_c1"]),
                "length": x,
                "from": "Posterior c1",
                "workchunk": str(workchunk),
            }
        )
        df_list.append(ndfp)
        ndfk = pd.DataFrame(
            {
                "density": length_dist(maxlle["weight_c2"]),
                "length": x,
                "from": "Posterior c2",
                "workchunk": str(workchunk),
            }
        )
        df_list.append(ndfk)
    if prior and (args := maxlle.get("args")):
        w_c1 = args.c1_dirichlet_prior
        y = length_dist(w_c1 / np.sum(w_c1))
        prior_df = pd.DataFrame(
            {"density": y, "length": x, "from": "Prior c1", "workchunk": "none",}
        )
        df_list.append(prior_df)
        w_c2 = args.c2_dirichlet_prior
        y = length_dist(w_c2 / np.sum(w_c2))
        prior_df = pd.DataFrame(
            {"density": y, "length": x, "from": "Prior c2", "workchunk": "none",}
        )
        df_list.append(prior_df)
    result_df = pd.concat(df_list)
    result_df["group"] = result_df["from"] + result_df["workchunk"]
    return result_df


def posterior_mode_weights(workdata, map_results):
    w_c1_posterior = w_c2_posterior = False
    for name, dat in workdata.iterrows():
        wg = dat["workchunk"]
        if wg not in map_results:
            continue
        n_cuts = len(dat["cuts"])
        w_c1_inferred = map_results[wg]["weight_c1"]
        w_c2_inferred = map_results[wg]["weight_c2"]
        if w_c1_posterior is False:
            w_c1_posterior = np.zeros(len(w_c1_inferred))
            w_c2_posterior = np.zeros(len(w_c2_inferred))
        w_c1_posterior += n_cuts * w_c1_inferred
        w_c2_posterior += n_cuts * w_c2_inferred
    return w_c1_posterior, w_c2_posterior


def check_length_distribution_flip(workdata, map_results, threshold=0.9):
    """Look for flip between posterior length distributions.

    :param threshold: Pearson correlation threshold for difference
        between c1 and c2 (default=.9).
    """
    logger.info("Calculating length distributions.")
    length_df = posterior_length_dists(map_results, prior=False)
    components = length_df["from"].unique()
    logger.info("Looking for flipped distributions.")
    diffs = dict()
    for wc, df in length_df.groupby("workchunk"):
        c1 = df[df["from"] == components[0]].set_index("length")["density"]
        c2 = df[df["from"] == components[1]].set_index("length")["density"]
        diffs[wc] = c1 - c2
    diffs = pd.DataFrame(diffs)

    usual_diff = diffs.median(axis=1)
    diff_coors = diffs.corrwith(usual_diff)
    idx_flipped = diff_coors < threshold
    bad_wgs = set(idx_flipped.index[idx_flipped].astype(int))

    if any(idx_flipped):
        w_c1_posterior, w_c2_posterior = posterior_mode_weights(workdata, map_results)
        wg_string = to_grouped_string(bad_wgs)
        post_str_c1 = " ".join([str(int(w)) for w in w_c1_posterior])
        post_str_c2 = " ".join([str(int(w)) for w in w_c2_posterior])
        logger.warn(
            """The length distributions for the following workgoups are flipped: %s
            Consider rerunning these workchunks with the following Dirichlet priors:

                --c1-dirichlet-prior %s
                --c2-dirichlet-prior %s

            To do that on slurm you can run

                jobs=%s
                sbatch --mem=[max memory] --array=$jobs sep241deconvolve.py \\
                    --c1-dirichlet-prior %s \\
                    --c2-dirichlet-prior %s \\
                    [additional parameters]
                """,
            wg_string,
            post_str_c1,
            post_str_c2,
            wg_string,
            post_str_c1,
            post_str_c2,
        )
    else:
        logger.info("No flipped length distributions found.")
    return bad_wgs


def read_job_data(jobdata_file):
    return pd.read_pickle(jobdata_file)


def to_grouped_string(list_of_integers):
    grouped_str = ""
    last_m = -np.inf
    in_group = False
    for m in sorted(list_of_integers):
        if m == last_m+1:
            in_group = True
        else:
            if in_group:
                # close of last group
                grouped_str += f"-{last_m}"
                in_group = False
            grouped_str += f",{m}"
        last_m = m
    if in_group:
        grouped_str += f"-{last_m}"
    grouped_str = grouped_str[1:] # remove leading ,
    return grouped_str


def read_results(jobdata_file, workdata, progress=True, error=True):
    map_results = dict()
    misses = list()
    last_wg_data = None
    for workchunk in tqdm(
        workdata["workchunk"].unique(), disable=~progress, desc="work chunk"
    ):
        filename = os.path.splitext(jobdata_file)[0] + f"_wg-{workchunk}.pkl"
        try:
            with open(filename, "rb") as fl:
                map_results[workchunk] = pickle.load(fl)
            last_wg_data = map_results[workchunk]
        except FileNotFoundError:
            misses.append(workchunk)
    if last_wg_data is None:
        raise NoData("No deconvolution results found.")
    if misses:
        total_wc = len(workdata["workchunk"].unique())
        miss_str = to_grouped_string(misses)
        logger.warn("Results for the following %s of %s work chunks are missing: %s",
                    f'{len(misses):,}', f'{total_wc:,}', miss_str)
        if error:
            raise MissingData(
                f"Results for the following {len(misses):,} of {total_wc:,} work chunks are missing: {miss_str}"
            )
    return map_results


class LevelSet:
    """
    Used by the gtf drawing tool.

    Remembers levels occupied by intervals if called
    for all intervals sorted by their start.
    Ensures a minimal padding distance between intervals in
    same level.
    """

    def __init__(self, interval_padding, base_level=0, increments=-2):
        self.padding = interval_padding
        self.base_level = base_level
        self.increments = increments
        self.occupied_levels = dict()

    def get_free(self, interval_start, interval_end):
        now_free = set()
        for level, end in self.occupied_levels.items():
            if end + self.padding < interval_start:
                now_free.add(level)
        [self.occupied_levels.pop(l) for l in now_free]
        level = self.base_level
        while level in self.occupied_levels.keys():
            level += self.increments
        self.occupied_levels[level] = interval_end
        return level

    def reset(self):
        self.occupied_levels = dict()


def igv_plot(
    df,
    color=None,
    group=None,
    facet_scales="free_y",
    min_counts=0,
    min_density=1e-10,
    sort_by=None,
    relabeling=dict(),
    smooth_events=None,
    max_res=1000,
    border=True,
):
    """
    Draws an IGV-like plot. Used by Multitool class.

    :param df: DataFrame with reads and columns 'start', 'stop' and
        columns specified by group and color.
    :param color: Column of df to color tracks by. Must be coarser than groups.
    :param group: Column of df to draw different tracks per group. Must be finer than color.
    :param facet_scales: passed to ggplot facet_grid to scale axis of tracks.
    :param min_count: Filter out tracks with less reads (default: 0).
    :param sort_by: Column of df. Sort tracks by the median within group. (default: 'rna_pseudotime')
    :param relabeling: Dict to rename track names.
    :param smooth_events: Plot kde of event rates instead of IGV if not None.
        Use value as bw_method in scipy.stats.gaussian_kde.
    :param max_res: Maximum grid resolution when plotting smooth_events.
    """
    import plotnine as p9
    if color is not None and group is None:
        group = color
    elif color is not None:
        if (df.groupby(group)[color].nunique() > 1).any():
            raise ValueError(
                f"More than one color per group. group: {group} color: {color}"
            )
    mdf = df.copy()
    if sort_by is not None and group is not None:
        sorted_cats = mdf.groupby(group)[sort_by].median().sort_values().index
        mdf[group] = pd.Categorical(mdf[group], categories=sorted_cats)
    if smooth_events is None:
        is_density = False
        sdf = mdf.sort_values("location")
        sdf["delta"] = np.where(sdf["variable"] == "start", 1, -1)
        if group is None:
            sdf["igv"] = sdf["delta"].cumsum()
        else:
            sdf["igv"] = sdf.groupby(group)["delta"].cumsum()
            if min_counts > 0:
                maxes = sdf.groupby(group)["igv"].transform(len) / 2
            if min_counts > 0:
                idx = maxes >= min_counts
                if not any(idx):
                    raise ValueError(
                        f"No group is above the min_counts of: {min_counts}"
                    )
                sdf = sdf.loc[idx, :]
        plot_df = sdf.copy()
        if group is None:
            plot_df["igv"] = plot_df["igv"].shift(1)
        else:
            plot_df["igv"] = plot_df.groupby(group)["igv"].shift(1)
        plot_df = plot_df.append(sdf)
        pl = (
            p9.ggplot(plot_df, p9.aes("location", ymin=0, ymax="igv"))
            + p9.geom_ribbon()
        )
    else:
        is_density = True
        lower_bound = mdf["location"].min() - 1
        upper_bound = mdf["location"].max() + 1
        grid = np.arange(lower_bound, upper_bound + 1)
        if max_res < len(grid):
            grid_idx = np.linspace(0, upper_bound - lower_bound, max_res).astype(int)
        else:
            grid_idx = slice(None)

        from KDEpy import FFTKDE
        if group is None:
            kernel = FFTKDE(kernel="gaussian", bw=smooth_events)
            kernel = kernel.fit(mdf["location"].values)
            values = kernel.evaluate(grid)
            factor = len(ldf["location"].values) * 100
            plot_df = pd.DataFrame(
                {
                    "density": kernel.evaluate(grid)[grid_idx],
                    "location": grid[grid_idx] * factor,
                }
            )
            if min_density is not None:
                plot_df.loc[plot_df["density"] < min_density, "density"] = min_density
        else:
            groups = mdf[group].unique()
            signals = list()
            for g in groups:
                idx = mdf[group] == g
                ldf = mdf.loc[idx, :]
                kernel = FFTKDE(kernel="gaussian", bw=smooth_events)
                kernel = kernel.fit(ldf["location"].values)
                values = kernel.evaluate(grid)[grid_idx]
                factor = len(ldf["location"].values) * 100
                signals.append(
                    pd.DataFrame(
                        {
                            "density": values * factor,
                            "location": grid[grid_idx],
                            group: g,
                            color: ldf[color].values[0],
                        }
                    )
                )
            plot_df = pd.concat(signals)
            if min_density is not None:
                plot_df.loc[plot_df["density"] < min_density, "density"] = min_density
            plot_df[group] = plot_df[group].astype(mdf[group].dtype)
            plot_df[color] = plot_df[color].astype(mdf[color].dtype)
        pl = p9.ggplot(
            plot_df, p9.aes("location", "density", ymax="density")
        ) + p9.geom_ribbon(ymin=0)
    if color is not None:
        if border:
            pl = pl + p9.aes(fill=color, color=color)
        else:
            pl = pl + p9.aes(fill=color)
    known_celltypes = {
        "HSC": "#F780BE",
        "HMP": "#E78AC2",
        "MEP": "#A7D75D",
        "Ery": "#4DAF50",
        "CLP": "#E5C397",
        "Mono": "#E50B22",
        "cDC": "#1A7AB2",
        "pDC": "#A5CFE2",
    }
    if color == "celltype" and all(
        ct in known_celltypes for ct in plot_df["celltype"].unique()
    ):
        # use familiar coloring
        (breaks, values) = zip(*known_celltypes.items())
        pl = pl + p9.scale_fill_manual(breaks=breaks, values=values,)
    else:
        pl = pl + p9.scale_fill_hue()

    if group is not None:
        pl += p9.facet_grid(
            f"{group} ~ .", scales=facet_scales, labeller=p9.labeller(relabeling),
        )
    if is_density:
        pl += p9.labs(y="cuts/hbp")
    else:
        pl += p9.labs(y="count")
    return pl


def plot_gtf(
    gtf_dataframe,
    draw_genes=True,
    draw_exons=True,
    draw_transcripts=True,
    draw_codons=True,
    draw_utr=True,
    draw_labels=True,
    padding=200,
):
    """
    Draw features of a gtf DataFrame. Used by Multitool.

    :param gtf_dataframe: A pandas data frame from a gtf parsed by gtfparse.
    :param draw_genes: Bool, if genes should be included.
    :param draw_exons: Bool, if exons should be drwan.
    :param draw_transcripts: Bool, if transcripts should be drawn (in a seperate track).
    :param draw_codons: Bool, if start and stop codon should be marked.
    :param draw_utr: Bool, if UTR should be drawn.
    :param draw_labels: Bool, if labels with gene names should be dran on genes.
    :param padding: Minimal horizontal distance in bp between drawn genes/transcripts.
        If features are too close, they a shifted vertically.
    """
    import plotnine as p9
    arrows = list()
    boxes = list()
    labels = list()
    levels = LevelSet(padding)
    if draw_genes:
        idx = gtf_dataframe["feature"] == "gene"
        genes_df = gtf_dataframe.loc[idx, :].sort_values("start")
        for gene, gene_dat in genes_df.iterrows():
            start = gene_dat["start"]
            end = gene_dat["end"]
            gene_name = gene_dat["gene_name"]
            level = levels.get_free(start, end)
            a_df = pd.DataFrame(
                {
                    "type": "gene",
                    "level": level,
                    "name": gene_name,
                    "start": gene_dat["stranded_start"],
                    "end": gene_dat["stranded_end"],
                    "feature": "interval",
                    "gene": gene_name,
                },
                index=[gene],
            )
            arrows.append(a_df)
            lab_df = pd.DataFrame(
                {
                    "type": "gene",
                    "level": level,
                    "name": gene_name,
                    "position": (gene_dat["start"] + gene_dat["end"]) / 2,
                    "gene": gene_name,
                },
                index=[gene],
            )
            labels.append(lab_df)
            most_exons = 1
            exon_hight = 0.6
            idx = gtf_dataframe["gene_name"] == gene_name
            gene_df = gtf_dataframe.loc[idx, :]
            if draw_exons:
                exon_entries = gene_df.loc[gene_df["feature"] == "exon", :]
                count_df = pd.concat(
                    [
                        pd.DataFrame({"location": exon_entries["start"], "delta": 1}),
                        pd.DataFrame({"location": exon_entries["end"], "delta": -1}),
                    ]
                )
                count_df = count_df.sort_values("location")
                signal = count_df["delta"].cumsum()
                count_df["box_up"] = level + exon_hight
                count_df["box_down"] = level - exon_hight
                most_exons = max(most_exons, signal.max())
                box_df = count_df.copy()
                box_df["start"] = box_df["location"]
                box_df["end"] = box_df["location"].shift(-1)
                box_df = box_df.loc[signal != 0, :]
                box_df["feature"] = "exon"
                box_df["type"] = "gene"
                box_df["gene"] = gene_name
                boxes.append(box_df)
            if draw_codons:
                hight = 1.5 * exon_hight
                starts = gene_df.loc[gene_df["feature"] == "start_codon", :].index
                for entry in starts:
                    codon_dat = gene_df.loc[entry, :]
                    box_df = pd.DataFrame(
                        {
                            "start": codon_dat["start"],
                            "end": codon_dat["end"],
                            "feature": "start codon",
                            "level": level,
                            "box_up": level + hight,
                            "box_down": level - hight,
                            "type": "gene",
                            "gene": gene_name,
                        },
                        index=[entry],
                    )
                    boxes.append(box_df)
                stops = gene_df.loc[gene_df["feature"] == "stop_codon", :].index
                for entry in stops:
                    codon_dat = gene_df.loc[entry, :]
                    box_df = pd.DataFrame(
                        {
                            "start": codon_dat["start"],
                            "end": codon_dat["end"],
                            "feature": "stop codon",
                            "level": level,
                            "box_up": level + hight,
                            "box_down": level - hight,
                            "type": "gene",
                            "gene": gene_name,
                        },
                        index=[entry],
                    )
                    boxes.append(box_df)
            if draw_utr:
                hight = 0.5 * exon_hight
                starts = gene_df.loc[gene_df["feature"] == "UTR", :].index
                for entry in starts:
                    utr_dat = gene_df.loc[entry, :]
                    box_df = pd.DataFrame(
                        {
                            "start": utr_dat["start"],
                            "end": utr_dat["end"],
                            "feature": "UTR",
                            "level": level,
                            "box_up": level + hight,
                            "box_down": level - hight,
                            "type": "gene",
                            "gene": gene_name,
                        },
                        index=[entry],
                    )
                    boxes.append(box_df)
    levels.reset()
    if draw_transcripts:
        idx = gtf_dataframe["feature"] == "transcript"
        tf_df = gtf_dataframe.loc[idx, :].sort_values("start")
        tf_shift = 0
        for trans in tf_df.index:
            trans_dat = tf_df.loc[trans, :]
            tf_name = trans_dat["transcript_id"]
            start = trans_dat["start"]
            end = trans_dat["end"]
            gene = trans_dat["gene_name"]
            level = levels.get_free(start, end)
            a_df = pd.DataFrame(
                {
                    "type": "transcript",
                    "level": level,
                    "name": tf_name,
                    "start": trans_dat["stranded_start"],
                    "end": trans_dat["stranded_end"],
                    "feature": "interval",
                    "gene": gene,
                },
                index=[trans],
            )
            arrows.append(a_df)
            if draw_exons:
                idx = (gene_df["feature"] == "exon") & (
                    gene_df["transcript_id"] == tf_name
                )
                exon_entries = gene_df.loc[idx, :]
                count_df = pd.concat(
                    [
                        pd.DataFrame({"location": exon_entries["start"], "delta": 1}),
                        pd.DataFrame({"location": exon_entries["end"], "delta": -1}),
                    ]
                )
                count_df = count_df.sort_values("location")
                signal = count_df["delta"].cumsum()
                signal = 0.8 * signal / signal.max()
                count_df["box_up"] = level + signal
                count_df["box_down"] = level - signal
                box_df = count_df.copy()
                box_df["start"] = box_df["location"]
                box_df["end"] = box_df["location"].shift(-1)
                box_df = box_df.loc[signal != 0, :]
                box_df["feature"] = "exon"
                box_df["type"] = "transcript"
                box_df["gene"] = gene
                boxes.append(box_df)
    pl = (
        p9.ggplot(p9.aes(fill="feature", color="feature"))
        + p9.facet_grid(f"type ~ .", scales="y_free")
        + p9.labs(x="location", y="tracks")
        + p9.theme(axis_text_y=p9.element_blank())
    )
    if boxes:
        boxes_df = pd.concat(boxes)
        pl = pl + (
            p9.geom_rect(
                boxes_df,
                p9.aes(xmin="start", xmax="end", ymin="box_down", ymax="box_up"),
                alpha=0.5,
            )
        )
    if arrows:
        arrow_df = pd.concat(arrows)
        pl = pl + p9.geom_segment(
            arrow_df,
            p9.aes(x="start", xend="end", y="level", yend="level"),
            size=1,
            arrow=p9.arrow(length=0.2),
        )
    if labels and draw_labels:
        labels_df = pd.concat(labels)
        pl = pl + p9.geom_label(
            labels_df,
            p9.aes("position", "level", label="name"),
            inherit_aes=False,
            alpha=0.7,
        )
    return pl


class DataManager:
    """
    A utility class for sep241 deconvolution.
    """

    def __init__(
        self, fragments_files=dict(), gtf_file=None, show_progress=True, comment="#",
        remove_pcr_duplicates = True,
    ):
        self.gtf_file = gtf_file
        self.fragments_files = fragments_files
        self.max_smooth_igv_resolution = 1000
        self._cache = dict()
        self.show_progress = show_progress
        self.comment = "#"
        self.remove_pcr_duplicates = remove_pcr_duplicates


    def plot_frag_length(self, log_scale=True):
        import plotnine as p9
        alen = self.all_fragments()
        pl = (
            p9.ggplot(df, p9.aes("lengths"))
            + p9.geom_line()
            + p9.ylim(0, None)
            + p9.ggtitle("atac read length distribution")
        )
        if log_scale:
            pl = pl + p9.scale_y_log10()
        return pl

    def plot_genes(self, genes, x="pseudotime", points=True, regression=True, std=True):
        import plotnine as p9
        if isinstance(genes, str):
            genes = genes.split(",")
        if regression:
            dfs = self.prediction(genes, x=x, std=std)
            rdf = pd.concat(dfs)
        if points:
            ad = self.rna_ad
            idx = ad.obs["selected"]
            df = pd.DataFrame({x: ad.obs[x][idx],}, index=ad.obs_names[idx],)
            for gene in genes:
                df[gene] = ad.layers["MAGIC_imputed_data"][idx, ad.var_names == gene]
            pdf = df.melt(id_vars=[x], var_name="gene", value_name="expression")
        pl = p9.ggplot(p9.aes(x, "expression", color="gene"))
        if regression and std:
            pl = pl + p9.geom_ribbon(
                p9.aes(x, ymin="lb", ymax="ub", fill="gene"),
                data=rdf,
                alpha=0.15,
                inherit_aes=False,
            )
        if points:
            pl = pl + p9.geom_point(data=pdf, alpha=0.05)
        if regression:
            pl = pl + p9.geom_line(data=rdf)
        return pl

    @property
    def gtf_file(self):
        return self._gtf_file

    @gtf_file.setter
    def gtf_file(self, gtf_file):
        if gtf_file is not None:
            logger.info("Parsing gtf file...")
            from gtfparse import read_gtf
            self.gtf = read_gtf(gtf_file)
        self._gtf_file = gtf_file

    @property
    def gtf(self):
        return self._gtf

    @gtf.setter
    def gtf(self, gtf):
        gtf["stranded_start"] = np.where(gtf["strand"] == "-", gtf["end"], gtf["start"])
        gtf["stranded_end"] = np.where(gtf["strand"] == "-", gtf["start"], gtf["end"])
        if not gtf["seqname"][0].startswith("chr"):
            gtf["seqname"] = 'chr' + gtf["seqname"]
        self.gene_names = set(gtf["gene_name"].unique())
        self._gtf = gtf

    def _interpret_feature(self, feature):
        try:
            return read_region_string(feature)
        except NotARegion:
            pass
        gene_gtf = self.gtf.loc[self.gtf["feature"] == "gene", :]
        if feature in self.gene_names:
            idx = np.where(gene_gtf["gene_name"] == feature)
            dat = gene_gtf.iloc[idx[0], :][["seqname", "start", "end"]]
            return dat.values.ravel()
        idx = np.where(gene_gtf["gene_id"] == feature)[0]
        if len(idx) > 0:
            dat = gene_gtf.iloc[idx[0], :][["seqname", "start", "end"]]
            return dat.values.ravel()
        idx = self.gtf["transcript_id"] == feature
        if any(idx):
            idx = np.where(idx & self.gtf["feature"] == "transcript")
            dat = self.gtf.iloc[idx[0], :][["seqname", "start", "end"]]
            return dat.values.ravel()
        raise Exception(f"Feature not recognized: {feature}")

    def get_fragments(
        self,
        feature,
        max_dist=0,
        include_events=True,
        nuc_size=120,
        _return_bounds=False,
        with_name=False,
    ):
        seq, start, end = self._interpret_feature(feature)
        if not seq.startswith("chr"):
            seq = f"chr{seq}"
        lower_bound = start - max_dist
        upper_bound = end + max_dist
        region = "%s:%d-%d" % (seq, lower_bound, upper_bound)
        atac_intervals = self.get_intervals(region, nuc_size=nuc_size, with_name=with_name)
        atac_intervals["seqname"] = seq
        if not include_events:
            return atac_intervals
        all_evens = self.envents_from_intervals(atac_intervals)
        if _return_bounds:
            return atac_intervals, all_evens, lower_bound, upper_bound
        return atac_intervals, all_evens

    def plot_igv(
        self,
        feature,
        max_dist=0,
        nuc_size=120,
        group=None,
        color="from",
        group_nucs=True,
        sort_by=None,
        obs_from=None,
        y_scale="free",
        draw_peaks=False,
        min_counts=0,
        min_density=1e-10,
        smooth_events_bw=None,
        mark_nucs=list(),
        anno_plot=True,
        color_anno_by="gene",
        full_label=False,
        igv_border=True,
        **kwargs,
    ):
        """
        A plot similar to the IGV visualization of ATAC reads.

        :param feature: A gene, genomic interval or df of ATAC reads.
        :param max_distance: Maximal distance around the selected feature to show.
        :param nuc_size: Length of ATAC reads at which to classify as nucleosome reads.
        :param group: Used in ggplot for faceting. Must be a subset of `color`\.
        :param color: Used in ggplot for coloring of tracks. Must be a superset of `group`\.
        :param group_nucs: Put nucleosome reads in a subgroup.
        :param sort_by: Sort groups and color by this column.
        :param obs_from: Select annData object. Must be `ATAC` or `RNA`\.
        :param y_scale: Must be `fixed` or `free`\.
        :param draw_peaks: Mark peaks selected in ATAC annData object.
        :param min_counts: Tracks with less than that will not be plotted.
        :param smooth_events_bw: If set to number draw event kde with that bw instead of igv signal.
        :param mark_nucs: List of nucleosome positions to mark.
        :param anno_plot: Call plot_gtf to show genes and genomic features.
        :param color_anno_by: E.g. `features` or `genes`\.
        :param full_label: Set to True to write full label into facet headers.
        :param \*\*kwargs: Args passed to plot_gtf.
        """
        if isinstance(feature, str):
            intervals, events, lower_bound, upper_bound = self.get_fragments(
                feature, max_dist=max_dist, nuc_size=nuc_size, _return_bounds=True,
            )
        else:
            events = feature
            max_dist = 0
            lower_bound = events["location"].min()
            upper_bound = events["location"].max()
        seq = events["seqname"][0]
        if y_scale == "free":
            facet_scales = "free_y"
        elif y_scale == "fixed":
            facet_scales = "fixed"
        else:
            raise ValueError(f"Argument for y_scale unknown: {y_scale}")

        if color in events.columns and group in events.columns:
            # we do not need to merge
            obs_from = None
        if obs_from is not None:
            if obs_from == "atac":
                obs = self.atac_ad.obs
            elif obs_from == "rna":
                obs = self.rna_ad.obs
            else:
                raise ValueError(f"Unrecognized obs_from: {obs_from}")
            events = pd.merge(
                events,
                obs,
                suffixes=("", obs_from),
                left_on="name",
                right_index=True,
                how="left",
            )
        if group is None and color is not None:
            group = color
        if min_counts > 0:
            idx = events.groupby(group)[group].transform(len) >= min_counts
            events = events.loc[idx, :]
        if sort_by is not None:
            if group is not None:
                presorted_group_cats = (
                    events.groupby(group)[sort_by].median().sort_values().index
                )
                if not group_nucs:
                    events[group] = pd.Categorical(
                        events[group], categories=presorted_group_cats
                    )
            if color is not None:
                sorted_cats = (
                    events.groupby(color)[sort_by].median().sort_values().index
                )
                events[color] = pd.Categorical(events[color], categories=sorted_cats)
        elif group is not None:
            presorted_group_cats = events[group].unique()

        relabeling = dict()
        if group_nucs:
            sorted_types = sorted(events["read_type"].unique(), reverse=True)
            if group is None:
                grouped = events["read_type"]
                sorted_group_cats = sorted_types
            else:
                grouped = events[group].astype(str) + " " + events["read_type"]
                sorted_group_cats = list()
                for ct in presorted_group_cats:
                    sorted_group_cats += [f"{ct} {t}" for t in sorted_types]
                    if full_label is not True:
                        relabeling.update({f"{ct} {t}": t for t in sorted_types})
            group = "_igv_group"
            events[group] = pd.Categorical(grouped, categories=sorted_group_cats)

        import plotnine as p9
        pl = igv_plot(
            events,
            group=group,
            color=color,
            sort_by=sort_by,
            facet_scales=facet_scales,
            min_counts=min_counts,
            min_density=min_density,
            relabeling=relabeling,
            smooth_events=smooth_events_bw,
            max_res=self.max_smooth_igv_resolution,
            border=igv_border,
        )
        pl = (
            pl
            + p9.coord_cartesian(xlim=(lower_bound, upper_bound))
            + p9.ggtitle(f"{seq}:{lower_bound:,}-{upper_bound:,}")
        )

        if draw_peaks:
            atac = self.atac_ad.var
            peaks_idx = (
                (atac["midpoint"] > lower_bound)
                & (atac["midpoint"] < upper_bound)
                & (atac["seqnames"] == seq)
            )
            atac = atac.loc[peaks_idx, :]
            pl = (
                pl
                + p9.geom_vline(
                    atac, p9.aes(xintercept="start"), linetype="solid", alpha=0.5
                )
                + p9.geom_vline(
                    atac, p9.aes(xintercept="end"), linetype="dashed", alpha=0.5
                )
            )
        if mark_nucs:
            shift = np.ceil(nuc_size / 2)
            nuc_starts = [mid - shift for mid in mark_nucs]
            nuc_ends = [p + nuc_size for p in nuc_starts]
            nucs = pd.DataFrame({"start": nuc_starts, "end": nuc_ends, "null": 0})
            pl = pl + p9.geom_segment(
                nucs,
                p9.aes("start", "null", xend="end", yend="null"),
                inherit_aes=False,
            )
        if not anno_plot:
            return pl

        # if seq.startswith('chr'):
        #    seq = seq[3:]
        idx = (
            (self.gtf["end"] > lower_bound)
            & (self.gtf["start"] < upper_bound)
            & (self.gtf["seqname"] == seq)
        )
        df = self.gtf.loc[idx, :]
        padding = (upper_bound - lower_bound) * 0.05
        kwargs.setdefault("padding", padding)
        kwargs.setdefault("draw_labels", True)
        kwargs.setdefault("draw_transcripts", False)
        ofs = p9.options.figure_size
        if kwargs.get("draw_transcripts"):
            vsize = 0.6 * ofs[0]
        else:
            vsize = 0.3 * ofs[0]
        if not any(idx):
            anno_pl = p9.ggplot()
        else:
            anno_pl = plot_gtf(df, **kwargs)
        anno_pl = (
            anno_pl
            + p9.theme(figure_size=(ofs[0], vsize))
            + p9.coord_cartesian(xlim=(lower_bound, upper_bound))
            + p9.aes(color=color_anno_by, fill=color_anno_by)
        )
        if draw_peaks:
            anno_pl = (
                anno_pl
                + p9.geom_vline(
                    atac, p9.aes(xintercept="start"), linetype="solid", alpha=0.5
                )
                + p9.geom_vline(
                    atac, p9.aes(xintercept="end"), linetype="dashed", alpha=0.5
                )
            )
        if mark_nucs:
            anno_pl = anno_pl + p9.geom_segment(
                nucs,
                p9.aes("start", "null", xend="end", yend="null"),
                inherit_aes=False,
            )
        return pl, anno_pl

    file_length = {}

    def _all_of_file(self, file, rep):
        comment = self.comment
        h = (
            ("f", "_all_of_file"),
            ("file", file),
            ("rep", rep),
            ("comment", comment),
        )
        if (result := self._cache.get(h)) is not None:
            return result
        reads = list()
        if file.endswith("gz"):
            logger.debug("Assuming gzip format for file %s", file)
            import gzip
            openfn = gzip.open
        elif file.endswith("gz2") or file.endswith("bz2"):
            logger.debug("Assuming bzip2 format for file %s", file)
            import bz2
            openfn = bz2.open
        else:
            logger.debug("Assuming no compression for file %s", file)
            openfn = open
        with openfn(file, "rb") as buff:
            absfile = os.path.abspath(file)
            if absfile in self.file_length:
                lines = tqdm(
                    buff,
                    desc="reads",
                    total=self.file_length[absfile],
                    disable=not self.show_progress,
                )
            else:
                lines = tqdm(buff, desc="reads", disable=not self.show_progress)
            n_fileds = None
            en_lines = enumerate(lines)
            for i, line in enumerate(lines):
                if n_fileds is None:
                    if not line.decode().strip().startswith(comment):
                        n_fileds = min(4, len(line.decode().split("\t")))
                        logger.debug("Using %d columns of %s.", n_fileds, file)
                if n_fileds == 3:
                    seq, s, e = line.decode().strip().split("\t")[:3]
                    start = int(s)
                    end = int(e)
                    reads.append((rep, seq, start, end, f"read-{i}"))
                elif n_fileds == 4:
                    seq, s, e, bc = line.decode().strip().split("\t")[:4]
                    start = int(s)
                    end = int(e)
                    reads.append((rep, seq, start, end, bc))
                elif line.decode().strip().startswith(comment):
                    continue
                else:
                    logger.error("Line %d of %s: %s", i, file, line)
                    raise BEDFormatError(f"Less than 3 columns in bed file {file}.")
        self._cache[h] = reads
        return reads

    def all_fragments(self, barcode=None, replace_barcode=None):
        reads = list()
        all_lengths = list()
        for rep, file in self.fragments_files.items():
            logger.info(f"Reading sample {rep}")
            reads += self._all_of_file(file, rep)
        logger.info("Making data frame.")
        result_df = pd.DataFrame(
            reads, columns=["from", "seqname", "start", "end", "name"]
        )
        if barcode:
            if barcode not in ["from", "name"]:
                raise BarcodeError("Barcode '%s' must be 'from' or 'name'.", barcode)
            result_df["barcode"] = result_df[barcode]
        total_reads = len(result_df)
        if replace_barcode:
            repl = lambda m: m.group(1)
            result_df['barcode'] = result_df['barcode'].str.replace(replace_barcode, repl, regex=True)
        if self.remove_pcr_duplicates:
            if "barcode" in result_df.columns:
                result_df = result_df.drop_duplicates(subset=["barcode", "seqname", "start", "end"])
            else:
                result_df = result_df.drop_duplicates()
            n_removed = total_reads - len(result_df)
            faction = n_removed / total_reads
            logger.info(f"Removed {n_removed:,} of {total_reads:,} ({faction:.2%}) duplicate reads.")
            total_reads = len(result_df)
        result_df["start"] = result_df["start"].astype(int)
        result_df["end"] = result_df["end"].astype(int)
        logger.info(f"Calculating read lengths for all {total_reads:,} reads.")
        result_df["length"] = result_df["end"] - result_df["start"]
        return result_df

    def get_intervals(self, region, nuc_size=120, with_name=False):
        intervals = list()
        import tabix
        for rep, fragments_file in self.fragments_files.items():
            tb = tabix.open(fragments_file)
            records = tb.querys(region)
            for record in records:
                if with_name:
                    intervals.append((rep, record[0], int(record[1]), int(record[2]), record[3]))
                else:
                    intervals.append((rep, record[0], int(record[1]), int(record[2])))
        if with_name:
            intervals_df = pd.DataFrame(intervals, columns=["from", "seq", "start", "end", "barcode"])
        else:
            intervals_df = pd.DataFrame(intervals, columns=["from", "seq", "start", "end"])
        intervals_df["length"] = np.abs(intervals_df["end"] - intervals_df["start"])
        intervals_df["read_type"] = np.where(
            intervals_df["length"] > nuc_size, f">{nuc_size}", f"≤{nuc_size}"
        )
        intervals_df["long_type"] = np.where(
            intervals_df["length"] > 2 * nuc_size,
            f">{2*nuc_size}",
            intervals_df["read_type"],
        )
        return intervals_df

    def envents_from_intervals(self, interval_df):
        id_vars = set(interval_df.columns) - {"start", "end"}
        df = interval_df.melt(id_vars=id_vars, value_name="location")
        return df

    def count(self, bins, index_name=None, normalize=False, count_prior=0):
        midpoints = (bins[1:] + bins[:-1]) / 2

        def do_count(data):
            cumhist = (data.values[None, :] > bins[:, None]).sum(axis=1)
            counts = np.diff(-cumhist)
            result = pd.Series(counts, index=midpoints)
            if index_name is not None:
                result.index.name = index_name
            result += count_prior
            if normalize:
                result /= result.sum()
            return result

        return do_count

    def _length_hist_bins(self, maximum, bin_width=1, padding=0.5):
        return np.arange(padding, maximum + padding + bin_width, bin_width)

    def hist_from_df(
        self,
        len_df,
        bin_width=1,
        normalize=True,
        color="from",
        count_prior=0,
        bins=None,
    ):
        if bins is None:
            bins = self._length_hist_bins(len_df["length"].max(), bin_width)
        cf = self.count(bins, "length", normalize=normalize, count_prior=count_prior)
        gdf = len_df.groupby(color)["length"].apply(cf)
        if normalize:
            gdf.name = "density"
        else:
            gdf.name = "count"
        return gdf

    def length_histograms(
        self, bin_width=1, normalize=True, color="from", count_prior=0, bins=None
    ):
        len_df = self.all_fragments()
        return self.hist_from_df(
            len_df,
            bin_width=bin_width,
            normalize=normalize,
            color=color,
            count_prior=count_prior,
            bins=bins,
        )

    def plot_length_hist(
        self,
        gdf=None,
        bin_width=1,
        normalize=True,
        color="from",
        count_prior=0,
        bins=None,
    ):
        import plotnine as p9
        if gdf is None:
            gdf = self.length_histograms(
                bin_width=bin_width,
                normalize=normalize,
                color=color,
                count_prior=count_prior,
                bins=bins,
            )
        ylab = gdf.name
        pdata = pd.DataFrame(gdf).reset_index()
        pl = (
            p9.ggplot(pdata, p9.aes("length", ylab, ymax=ylab, color=color, fill=color))
            + p9.geom_ribbon(ymin=0, alpha=0.1)
            + p9.ggtitle(f"length bin width {bin_width}")
        )
        return pl



import os
import re
import pickle

from threadpoolctl import threadpool_limits
import numpy as np
import pandas as pd
import multiprocessing

import theano
import pymc3 as pm
import theano.tensor as at

from sep241util import setup_logging, logger, format_cuts
from sep241covariance import SparseCov
from sep241latent import SparseLatent
from pymc3.theanof import inputvars
from pymc3.model import Point, modelcontext
from pymc3.blocking import DictToArrayBijection, ArrayOrdering
from scipy.optimize import Bounds


class NoData(Exception):
    pass


class ArgumentError(Exception):
    pass


class BadCovariance(Exception):
    pass


def safety_check(code):
    """
    Checks that the code is safe to execute. If the code is dangerous, raises
    an exception. The converse is not true. If the code is a valid covariance
    function, does not raise an exception. The converse is not true.
    """
    free = "([ |\(|\)]*)"
    dec = f"({free}((\.[0-9]+)|([0-9]+(\.[0-9]+)?)){free})"
    cov = f"({free}(pm.gp.cov.[0-9a-zA-Z]*\({dec}(,{dec})*\)){free})"
    values = f"({dec}|{cov})"
    operators = "(\+|\*|\*\*)"
    expression = f"{values}({operators}{values})*"

    if not re.fullmatch(expression, code):
        raise BadCovariance()


def get_cov_functions(c1_cov_string, c2_cov_string, threshold):
    p = re.compile(r"([a-zA-Z][a-zA-Z0-9]*\()")
    c1_code = p.sub(r"pm.gp.cov.\g<1>1, ", c1_cov_string)
    c2_code = p.sub(r"pm.gp.cov.\g<1>1, ", c2_cov_string)
    covs_wo_input_dims = {
        "WhiteNoise(": r"WhiteNoise\(1, ",
        "Constant(": r"Constant\(1, ",
        "Kron(": r"Kron\(1, ",
    }
    for repl, pattern in covs_wo_input_dims.items():
        c1_code = re.sub(pattern, repl, c1_code)
        c2_code = re.sub(pattern, repl, c2_code)
    try:
        safety_check(c1_code)
        c1_cov = eval(c1_code)
    except Exception:
        raise BadCovariance(
            "Error parsing the covariance function string "
            f'for component 1: "{c1_cov_string}"'
        )
    try:
        safety_check(c2_code)
        c2_cov = eval(c2_code)
    except Exception:
        raise BadCovariance(
            "Error parsing the covariance function string "
            f'for component 2: "{c2_cov_string}"'
        )

    wrap_1 = SparseCov(c1_cov, threshold=threshold)
    wrap_2 = SparseCov(c2_cov, threshold=threshold)
    return [wrap_1, wrap_2]


def get_length_dist_modes(modes=[70, 200, 400, 600], sigmas=[0.29, 0.18, 0.15, 0.085]):
    def generator():
        mus = [np.log(modes[i]) + (s ** 2) for i, s in enumerate(sigmas)]
        length_comps = list()
        for i in range(len(mus)):
            mode = pm.Lognormal.dist(mus[i], sigma=sigmas[i], testval=modes[i])
            length_comps.append(mode)
        return length_comps

    return generator


def make_bounds():
    k = 9.458853214586302
    model = modelcontext(None)
    lpoint = Point(model.test_point, model=model)
    upoint = Point(model.test_point, model=model)
    for name, value in lpoint.items():
        if name == "weight_c1_stickbreaking__" or name == "weight_c2_stickbreaking__":
            lpoint[name] = np.full(value.shape, -k)
            upoint[name] = np.full(value.shape, k)
        else:
            lpoint[name] = np.full(value.shape, -np.inf)
            upoint[name] = np.full(value.shape, np.inf)

    vars = model.cont_vars
    vars = inputvars(vars)

    lbij = DictToArrayBijection(ArrayOrdering(vars), lpoint)
    ubij = DictToArrayBijection(ArrayOrdering(vars), upoint)

    lb = lbij.map(lpoint)
    ub = ubij.map(upoint)
    return Bounds(lb, ub, True)


def make_model(
    events,
    cov_functions,
    dirichlet_priors,
    length_comps_gen,
    mlevel=0,
    constrain=False,
):
    logger.info("Compiling model.")
    with pm.Model() as model:
        length_comps = length_comps_gen()

        cov_c1 = cov_functions[0]
        gp_c1 = SparseLatent(
            mean_func=pm.gp.mean.Constant(mlevel),
            cov_func=cov_c1,
        )
        w_c1 = dirichlet_priors[0]
        weights_c1 = pm.Dirichlet("weight_c1", w_c1, testval=w_c1 / np.sum(w_c1))
        l_like_c1 = pm.Mixture.dist(w=weights_c1, comp_dists=length_comps[: len(w_c1)])

        cov_c2 = cov_functions[1]
        gp_c2 = SparseLatent(
            mean_func=pm.gp.mean.Constant(mlevel),
            cov_func=cov_c2,
        )
        w_c2 = dirichlet_priors[1]
        weights_c2 = pm.Dirichlet("weight_c2", w_c2, testval=w_c2 / np.sum(w_c2))
        l_like_c2 = pm.Mixture.dist(w=weights_c2, comp_dists=length_comps[: len(w_c2)])

        if constrain:
            logger.debug("Compiling constrain.")
            means = at.stack([a.mean for a in length_comps])
            mean_prior_length_c1 = means.dot(w_c1 / np.sum(w_c1)).eval()
            mean_prior_length_c2 = means.dot(w_c2 / np.sum(w_c2)).eval()
            if mean_prior_length_c1 > mean_prior_length_c2:
                logger.info(
                    "Constraining component 1 to have larger fragments on average."
                )
                larger_weights = weights_c1
                smaller_weights = weights_c2
            else:
                logger.info(
                    "Constraining component 2 to have larger fragments on average."
                )
                larger_weights = weights_c2
                smaller_weights = weights_c1
            length_diff = means.dot(larger_weights) - means.dot(smaller_weights)
            const = pm.Potential(
                "length_mode_constrain",
                pm.math.switch(length_diff > 0, 0, -length_diff * 1e99),
            )

        n_intervals = len(events)
        events["par_work"] = events["nlocs"].astype(float) * events["ncuts"].astype(
            float
        )
        events = events.sort_values(by="par_work", ascending=False)
        for i, (name, dat) in enumerate(events.iterrows()):
                unique_locations, log_loc_weight, interleave_index = format_cuts(
                    dat["cuts"]["location"]
                )
                lengths = dat["cuts"]["length"].values
                logger.info(
                    "Parameterizing interval (%d/%d) %s with %i cuts in %i locations.",
                    i + 1,
                    n_intervals,
                    name,
                    len(dat["cuts"]),
                    len(unique_locations),
                )

                f_c1 = gp_c1.prior(f"f_c1_{name}", X=unique_locations[:, None])
                logp_c1 = f_c1[interleave_index] + l_like_c1.logp(lengths)

                f_c2 = gp_c2.prior(f"f_c2_{name}", X=unique_locations[:, None])
                logp_c2 = f_c2[interleave_index] + l_like_c2.logp(lengths)

                comb = at.stack([logp_c2, logp_c1], axis=0)
                comb_logp = pm.math.logsumexp(comb, axis=0)
                potential = pm.Potential(f"potential_{name}", comb_logp)

                weight_c1 = at.sum(at.exp(f_c1 + log_loc_weight))
                weight_c2 = at.sum(at.exp(f_c2 + log_loc_weight))
                weight_total = weight_c1 + weight_c2
                n_cuts = len(dat["cuts"])
                t_potential = pm.Potential(
                    f"total_{name}",
                        pm.Normal.dist(np.log(n_cuts), 0.001).logp(at.log(weight_total)),
                )

    return model


def gradient1_custom(f, v):
    theano.config.compute_test_value = "off"
    result = at.flatten(at.grad(f, v, disconnected_inputs="warn"))
    return result


def deconvolve(args):
    setup_logging(args.logLevel, args.logfile)
    logger.info("Running whole_genome_deconvolve_job with arguments:")
    logger.info(args)
    logger.debug("Theano flags: %s", os.getenv("THEANO_FLAGS"))

    if not args.cores:
        cores = multiprocessing.cpu_count()
        logger.info("Detecting %s compute cores.", cores)
    else:
        cores = args.cores
    logger.info("Theano compile directory set to %s", theano.config.compiledir)
    logger.info("Limiting computation to %i cores.", cores)
    threadpool_limits(limits=int(cores))
    os.environ["NUMEXPR_MAX_THREADS"] = str(cores)
    logger.info("Preparing dirichlet prior.")
    c1_prior = np.array(args.c1_dirichlet_prior)
    if len(c1_prior) < 2:
        raise ArgumentError("Dirichlet prior of c1 needs at least two components.")
    nmodes = len(args.length_dist_modes)
    if len(args.length_dist_mode_sds) != nmodes:
        raise ArgumentError(
            "Number of fragment length distribution mode standard deviations must "
            "equal the number of fragment length distribution modes."
        )
    if len(c1_prior) > nmodes:
        raise ArgumentError(
            "Dirichlet prior of c1 has more than components than modes defined "
            "for the fragment length distribution."
        )
    c2_prior = np.array(args.c2_dirichlet_prior)
    if len(c2_prior) < 2:
        raise ArgumentError("Dirichlet prior of c2 needs at least two components.")
    if len(c2_prior) > nmodes:
        raise ArgumentError(
            "Dirichlet prior of c2 has more than components than modes defined "
            "for the fragment length distribution."
        )
    dirichlet_priors = [c1_prior, c2_prior]
    logger.info("Making covariance functions.")
    cov_functions = get_cov_functions(args.c1_cov, args.c2_cov, args.sparsity_threshold)
    logger.info("Loading workdata from: %s", args.jobdata)
    workdata = pd.read_pickle(args.jobdata)
    logger.info("Selecting intervals of work chunk %s.", args.workchunk)
    idx = workdata["workchunk"].astype(str) == args.workchunk
    events = workdata.loc[idx, :]
    del workdata
    if len(events) == 0:
        raise NoData(f"No intervals found for work chunk {args.workchunk}.")
    workload = events["memory"].sum() + 9.4640381e-2
    logger.info(
        "Found %i intervals with a target memory consumption of %s GBs.",
        len(events),
        f"{workload:,.3f}",
    )
    length_comps_gen = get_length_dist_modes(
        args.length_dist_modes, args.length_dist_mode_sds
    )
    model = make_model(
        events,
        cov_functions,
        dirichlet_priors,
        length_comps_gen,
        constrain=args.constrain,
    )

    logger.info("Computing MAP.")
    with model:
        pm.theanof.gradient1 = gradient1_custom
        maxlle = pm.find_MAP(bounds=make_bounds())
    out_path = os.path.splitext(args.jobdata)[0] + f"_wg-{args.workchunk}.pkl"
    logger.info("Saving result in %s.", out_path)
    maxlle["arguments"] = args
    with open(out_path, "wb") as fl:
        pickle.dump(maxlle, fl, protocol=pickle.HIGHEST_PROTOCOL)
    logger.info("Finished successfully.")


#!/usr/bin/env python
#SBATCH --job-name=2for1separator
#SBATCH --cpus-per-task=1
#SBATCH --nodes=1
#SBATCH --ntasks-per-node=1
#SBATCH --time=2-00:00:00


import os
import argparse

from sep241.sep241util import set_flag


desc = """Deconvolve CUT&TAG 2for1. Uses the output of the sep241prep step.

This step computes two likelihoods for all cuts (both ends of each fragment) for
them to originate from one of two channels. By default, we expect that
channel 1 contains smaller fragments, e.g., Pol2S5p induced, than channel 2,
e.g., H3K27me3 induced. This is controlled by ``--c1-dirichlet-prior`` and
``--c2-dirichlet-prior``. The dirichlet prior specifies weights for multiple
log-normal distributions that make up a mixture distribution describing the prior
fragment-length distribution that typically asserts a ladder of distinct modes.

The mode position and width is shared between both channels and controlled
by ``--length-dist-modes`` and ``--length-dist-mode-sds``.
The default setting further assumes that co-occurrence of cuts in channel 1 is
more concentrated than channel 2, in which cuts are assumed to be more spread
out. This is controlled through the covariance functions of the respective
position marginal log-likelihood functions ``--c1-cov`` and ``--c2-cov``.

A single instance of the deconvolution process deconvolves one workchunk
specified by ``--workchunk``. Either of the following code snippets run all the
deconvolutions at once, where each variable with args. is an argument
to the deconvolution step and ``[N]`` is the highest workchunk id.

Run deconvolution using `slurm <https://slurm.schedmd.com>`_::

  sbatch --array=0-[N] --mem=[memory target] sep241deconvolve [jobdata pkl file]

Run in bash or zsh::

  for wc in $(seq 0 [N]); do
      sep241deconvolve [jobdata pkl file] --workchunk $wc
  done

The preparation step tries to create workgroups that will take the
target memory demand, but this is only an estimate. If the memory demand
of the deconvolution step is prohibitively high, try rerunning the
preparation step with a lower memory target.
"""

parser = argparse.ArgumentParser(description=desc)
parser.add_argument(
    "jobdata",
    metavar="jobdata-file",
    type=str,
    nargs="?",
    help="Jobdata with cuts per interval and workchunk ids.",
)
parser.add_argument(
    "-l",
    "--log",
    dest="logLevel",
    choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    default="INFO",
    help="Set the logging level (default=INFO)",
    metavar="LEVEL",
)
parser.add_argument(
    "--logfile",
    help="Write detailed log to this file.",
    type=str,
    metavar="logfile",
)
parser.add_argument(
    "--workchunk",
    help="""Work chunk number of jobdata to handle.
    Defaults to the environment variable SLURM_ARRAY_TASK_ID if set.""",
    type=str,
    default=os.getenv("SLURM_ARRAY_TASK_ID"),
    metavar="int",
)
parser.add_argument(
    "--c1-cov",
    help="""Covariance function of component 1 using pymc
    covariance functions without the `input_dim` argument:
    https://docs.pymc.io/api/gp/cov.html
    (default=Matern32(500))
    """,
    type=str,
    default="Matern32(500)",
    metavar="'a*CovFunc(args) + ...'",
)
parser.add_argument(
    "--c2-cov",
    help="""Covariance function of component 2 using pymc
    covariance functions without the `input_dim` argument:
    https://docs.pymc.io/api/gp/cov.html
    (default=Matern32(2000))
    """,
    type=str,
    default="Matern32(2000)",
    metavar="'a*CovFunc(args) + ...'",
)
parser.add_argument(
    "--sparsity-threshold",
    help="""The Gaussian processes modeling the location distribution use a custom
    covariance matrix that saves time and memory by representing only
    covariances above this threshold.
    (default=1e-8)""",
    type=float,
    default=1e-8,
    metavar="float",
)
parser.add_argument(
    "--length-dist-modes",
    help="""Modes of the log-normal distributions used in the
    mixture model for the fragment length distribution.
    (default=70 200 400 600)
    """,
    type=float,
    nargs="+",
    default=[70, 200, 400, 600],
    metavar="floats",
)
parser.add_argument(
    "--length-dist-mode-sds",
    help="""Standard deviations of the log-normal modes of
    the mixture model for the fragment length distribution.
    (default=0.29 0.18 0.15 0.085)
    """,
    type=float,
    nargs="+",
    default=[0.29, 0.18, 0.15, 0.085],
    metavar="floats",
)
parser.add_argument(
    "--c1-dirichlet-prior",
    help="""Dirichlet prior for the ratio of modes in the
    length distribution.
    (default=450 100 10 1)
    """,
    type=float,
    nargs="+",
    default=[450, 100, 10, 1],
    metavar="floats",
)
parser.add_argument(
    "--c2-dirichlet-prior",
    help="""Dirichlet prior for the ratio of modes in the
    length distribution.
    (default=150 300 50 10)
    """,
    type=float,
    nargs="+",
    default=[150, 300, 50, 10],
    metavar="floats",
)
parser.add_argument(
    "--constrain",
    help="""Use this flag to constrain average fragment length to be larger in one component.
    (default=False)""",
    action="store_true",
)
parser.add_argument(
    "--cores",
    help="""Number of CPUs to use for inference.
    Defaults to the environment variable SLURM_CPUS_PER_TASK if set.""",
    type=int,
    default=os.getenv("SLURM_CPUS_PER_TASK"),
    metavar="int",
)
parser.add_argument(
    "--compiledir",
    help="""Directory to compile code in.
    Should be different between parallel run instances
    and can be deleted after run.
    Defaults to sep241tmp/{args.jobdata}/deconvolve/{args.workchunk} or if the environment variable TMPDIR is set, defaults to TMPDIR/sep241tmp/{args.jobdata}/deconvolve/{args.workchunk}""",
    type=str,
    metavar="dir",
)


def main():
    args = parser.parse_args()
    tmpdir = os.getenv("TMPDIR")
    if args.compiledir is not None:
        compileDir = args.compiledir
    elif tmpdir is None:
        compileDir = f"./sep241tmp/{args.jobdata}/deconvolve/{args.workchunk}"
    else:
        compileDir = os.path.join(tmpdir, f"sep241tmp/{args.jobdata}/deconvolve/{args.workchunk}")
    set_flag("base_compiledir", compileDir)
    from sep241.sep241model import deconvolve
    deconvolve(args)


if __name__ == "__main__":
    main()



#! /usr/bin/env python
import os
import argparse
import pickle

import numpy as np
import scipy
import pandas as pd
from tqdm.auto import tqdm
from threadpoolctl import threadpool_limits
import multiprocessing

from sep241util import logger, setup_logging
from sep241util import read_job_data, read_results, write_bed, detect_cores
from sep241util import check_length_distribution_flip
from sep241util import MissingData


desc = """Peak calling after CUT&TAG 2for1 deconvolution.
We use the positional marginal likelihoods to originate from one of the two
channels to identify peaks of high cut likelihood for the respective channels
according to
`CUT&Tag2for1: a modified method
for simultaneous profiling of the accessible and silenced regulome in single
cells <https://doi.org/10.1186/s13059-022-02642-w>`_.

By default, channel 1 is assumed to contain narrow, sharp peaks. Channel 2
is assumed to contain broad domains of a more gradually changing signal. The minimum
peak size can be controlled by ``--c1-min-peak-size`` and
``--c2-min-peak-size``. Additionally, signals can be
smoothed through ``--c1-smooth`` and ``--c2-smooth`` before peak calling
to reflect a more gradual change.
"""
parser = argparse.ArgumentParser(description=desc)
parser.add_argument(
    "jobdata",
    metavar="jobdata-file",
    type=str,
    nargs="?",
    help="Jobdata with cuts per interval and workchunk ids.",
)
parser.add_argument(
    "-l",
    "--log",
    dest="logLevel",
    choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    default="INFO",
    help="Set the logging level (default=INFO).",
    metavar="LEVEL",
)
parser.add_argument(
    "--logfile",
    help="Write detailed log to this file.",
    type=str,
    metavar="logfile",
)
parser.add_argument(
    "-o",
    "--out",
    help="Output directory (default is the directory of the jobdata input file).",
    type=str,
    metavar="out_dir",
)
parser.add_argument(
    "--c1-min-peak-size",
    help="Minimal number of bases per peak for channel 1 (default=100).",
    type=int,
    default=100,
    metavar="int",
)
parser.add_argument(
    "--c2-min-peak-size",
    help="Minimal number of bases per peak for channel 2 (default=400).",
    type=int,
    default=400,
    metavar="int",
)
parser.add_argument(
    "--c1-smooth",
    help="Apply gaussian filter with this standard deviation to channel 1 (default=0).",
    type=float,
    default=0,
    metavar="float",
)
parser.add_argument(
    "--c2-smooth",
    help="Apply gaussian filter with this standard deviation to channel 2 (default=2000).",
    type=float,
    default=2000,
    metavar="float",
)
parser.add_argument(
    "--fraction-in-peaks",
    help="Fraction of reads that are expected to be in peaks (default=0.5).",
    type=float,
    default=0.5,
    metavar="float",
)
parser.add_argument(
    "--fraction-overlap",
    help="""If more than this fraction of a peak overlaps with a peak of the
    other target, then it is considered an overlapping region (default=0.5).""",
    type=float,
    default=0.5,
    metavar="float",
)
parser.add_argument(
    "--span",
    help="Resolution in number of base pairs (default=10).",
    type=int,
    default=10,
    metavar="int",
)
parser.add_argument(
    "--no-check",
    help="Do not test for flipped length distributions.",
    action="store_true",
)
parser.add_argument(
    "--only-check", 
    help="Stop after initial checking.",
    action="store_true",
)
parser.add_argument(
    "--uncorrected",
    help="Do not correct cut ratio estimate with Bayesian prior.",
    action="store_true",
)
parser.add_argument(
    "--force",
    help="Call peaks even if some results are missing.",
    action="store_true",
)
parser.add_argument(
    "--no-progress", help="Do not show progress.", action="store_true",
)
parser.add_argument(
    "--cores",
    help="Number of CPUs to use for the preparation.",
    type=int,
    default=0,
    metavar="int",
) 


def interpolate_entry(job):
    (
        interval_name,
        seqname,
        locations,
        signal_c1,
        c1_sigma,
        signal_c2,
        c2_sigma,
        step,
    ) = job
    grid = np.arange(np.min(locations), np.max(locations) + 1)
    interpolated_c1 = np.interp(grid, locations, signal_c1)[::step]
    interpolated_c2 = np.interp(grid, locations, signal_c2)[::step]
    df = pd.DataFrame(
        {
            "interval": interval_name,
            "seqname": seqname,
            "location": grid[::step],
            "c1": interpolated_c1,
            "c2": interpolated_c2,
        }
    )
    if c1_sigma:
        df["c1 smooth"] = scipy.ndimage.gaussian_filter1d(
            interpolated_c1, int(c1_sigma / step)
        )
    if c2_sigma:
        df["c2 smooth"] = scipy.ndimage.gaussian_filter1d(
            interpolated_c2, int(c2_sigma / step)
        )
    return df


def make_interpolation_jobs(workdata, map_results, c1_sigma, c2_sigma, step_size):
    jobs = list()
    signal_c1_list = list()
    signal_c2_list = list()
    signal_c1_list_global = list()
    signal_c2_list_global = list()
    location_list = list()
    last_name = ""

    max_log_value = None
    missing_data_wg = set()

    for name, dat in tqdm(workdata.iterrows(), total=len(workdata), desc="intervals"):
        wg = dat["workchunk"]
        maxlle = map_results.get(dat["workchunk"], None)
        if maxlle is None:
            if wg not in missing_data_wg:
                logger.warning("Data of workchunk %d is None.", wg)
                missing_data_wg.add(wg)
            continue
        if f"f_c1_{name}" not in maxlle.keys():
            logger.warning("Marginal likelihood results not found in workchunk %d.", wg)
            continue
        if max_log_value is None:
            max_log_value = np.log(np.finfo(maxlle[f"f_c1_{name}"].dtype).max)
        start_idx = dat["stich_start"][0]
        end_idx = dat["stich_end"][0]
        seq = dat["seqname"]
        locations = dat["cuts"]["location"][start_idx:end_idx].values.astype(int)

        idx = dat["cuts"]["location"].rank(method="dense").astype(int) - 1

        log_signal_c1 = maxlle[f"f_c1_{name}"][idx][start_idx:end_idx]
        log_signal_c2 = maxlle[f"f_c2_{name}"][idx][start_idx:end_idx]
        if np.max(log_signal_c1) > max_log_value:
            logger.warning(
                f"The c1 track of interval {name} in work chunk {wg} contains "
                "values that are too large. The interval will be skipped."
            )
            continue
        if np.max(log_signal_c2) > max_log_value:
            logger.warning(
                f"The c2 track of interval {name} in work chunk {wg} contains "
                "values that are too large. The interval will be skipped."
            )
            continue
        signal_c1 = np.exp(log_signal_c1)
        signal_c2 = np.exp(log_signal_c2)

        signal_c1_list_global.append(signal_c1)
        signal_c2_list_global.append(signal_c2)
        base_name, _ = name.split(".")
        if location_list and last_name != base_name:
            jobs.append(
                (
                    last_name,
                    last_seq,
                    np.hstack(location_list),
                    np.hstack(signal_c1_list),
                    c1_sigma,
                    np.hstack(signal_c2_list),
                    c2_sigma,
                    step_size,
                )
            )
            location_list = list()
            signal_c1_list = list()
            signal_c2_list = list()
        if dat["is_subdivide"]:
            location_list.append(locations)
            signal_c1_list.append(signal_c1)
            signal_c2_list.append(signal_c2)
            last_seq = seq
            last_name = base_name
        else:
            jobs.append(
                (
                    base_name,
                    seq,
                    locations,
                    signal_c1,
                    c1_sigma,
                    signal_c2,
                    c2_sigma,
                    step_size,
                )
            )
    return jobs, signal_c1_list_global, signal_c2_list_global


def interpolate(
    workdata, map_results, c1_sigma, c2_sigma, step_size, cores=8, progress=False
):
    logger.info("Splitting data into intervals.")
    jobs, signal_c1_list_global, signal_c2_list_global = make_interpolation_jobs(
        workdata, map_results, c1_sigma, c2_sigma, step_size
    )
    results_list = list()
    if c1_sigma or c2_sigma:
        if not c1_sigma:
            smooth_msg = "c2"
        elif not c1_sigma:
            smooth_msg = "c1"
        else:
            smooth_msg = "c1 and c2"
    else:
        smooth_msg = "none"
    logger.info("Interpolating + smoothening %s.", smooth_msg)
    n_jobs = len(jobs)
    mutliprocessing_chunksize = min(10, int(np.ceil(n_jobs/cores)))
    with threadpool_limits(limits=1):
        with multiprocessing.Pool(cores) as pool:
            for df in tqdm(
                pool.imap(interpolate_entry, jobs, mutliprocessing_chunksize),
                total=n_jobs,
                disable=not progress,
                desc="intervals",
            ):
                results_list.append(df)
    comb_data = pd.concat(results_list, axis=0)
    comb_data.index = range(len(comb_data))
    return comb_data, signal_c1_list_global, signal_c2_list_global


def target_fractions(comb_data, fraction_in_peaks, uncorrected=False):
    total_c1 = comb_data["c1"].sum()
    total_c2 = comb_data["c2"].sum()
    total = total_c1 + total_c2
    fraction_c1_uncorrected = total_c1 / total
    fraction_c2_uncorrected = total_c2 / total
    logger.info(
        f"Integral ratio: {fraction_c1_uncorrected:.1%} c1 and {fraction_c2_uncorrected:.1%} c2 cuts."
    )
    if uncorrected:
        fraction_c1 = fraction_in_peaks * total_c1 / total
        fraction_c2 = fraction_in_peaks * total_c2 / total

    equal_dist_cuts = total
    total_c1 = comb_data["c1"].sum() + (equal_dist_cuts / 2)
    total_c2 = comb_data["c2"].sum() + (equal_dist_cuts / 2)
    total = total_c1 + total_c2
    fraction_c1_permissive = total_c1 / total
    fraction_c2_permissive = total_c2 / total
    logger.info(
        f"Bayesian estimation: {fraction_c1_permissive:.1%} c1 and {fraction_c2_permissive:.1%} c2 cuts."
    )

    if uncorrected:
        logger.info(
            f"Estimating an uncorrected total of {fraction_c1:.1%} c1 and {fraction_c2:.1%} c2 cuts in peaks."
        )
        return (
            fraction_c1,
            fraction_c2,
            fraction_c1_uncorrected,
            fraction_c2_uncorrected,
        )

    fraction_c1 = fraction_in_peaks * total_c1 / total
    fraction_c2 = fraction_in_peaks * total_c2 / total
    logger.info(
        f"Estimating a total of {fraction_c1:.1%} c1 and {fraction_c2:.1%} c2 cuts in peaks."
    )
    return fraction_c1, fraction_c2, fraction_c1_permissive, fraction_c2_permissive


def estimated_deconvolved_fraction(workdata):
    total = 0
    covered = 0
    for seqname, data in workdata.groupby("seqname"):
        total += np.max(data["end"])
        covered += np.sum(data["end"] - data["start"])
    return covered / total


def mark_peaks(
    comb_data,
    signal_c1_list_global,
    signal_c2_list_global,
    fraction_c1,
    fraction_c2,
    work_coverage=None,
):

    bound_c1 = np.quantile(np.hstack(signal_c1_list_global), 1 - fraction_c1)
    if "c1 smooth" in comb_data.keys():
        logger.info("Averaging original and smoothed signal of c1.")
        comb_data["c1 mean"] = np.mean(
            comb_data.loc[:, ["c1", "c1 smooth"]].values, axis=1
        )
        comb_data["c1_peak"] = (comb_data["c1 mean"] > bound_c1) | (
            comb_data["c1 smooth"] > bound_c1
        )
    else:
        comb_data["c1_peak"] = comb_data["c1"] > bound_c1
    fraction_selected = comb_data["c1_peak"].sum() / len(comb_data["c1_peak"])
    msg = f"c1: {fraction_selected:.2%} of intervals"
    if work_coverage:
        msg += f" and {work_coverage*fraction_selected:.2%} of genome selected."
    logger.info(msg)

    bound_c2 = np.quantile(np.hstack(signal_c2_list_global), 1 - fraction_c2)
    if "c2 smooth" in comb_data.keys():
        logger.info("Averaging original and smoothed signal of c2.")
        comb_data["c2 mean"] = np.mean(
            comb_data.loc[:, ["c2", "c2 smooth"]].values, axis=1
        )
        comb_data["c2_peak"] = (comb_data["c2 mean"] > bound_c2) | (
            comb_data["c2 smooth"] > bound_c2
        )
    else:
        comb_data["c2_peak"] = comb_data["c2"] > bound_c2
    fraction_selected = comb_data["c2_peak"].sum() / len(comb_data["c2_peak"])
    msg = f"c2: {fraction_selected:.2%} of intervals"
    if work_coverage:
        msg += f" and {work_coverage*fraction_selected:.2%} of genome selected."
    logger.info(msg)
    return comb_data


def track_to_interval(loc_comb_data, comp_name, step_size, seqname):
    idx = np.insert(loc_comb_data[f"{comp_name}_peak"].values.astype(int), 0, 0)
    indicator = np.diff(idx)
    loc_comb_data[f"{comp_name}_peak_number"] = np.cumsum(indicator == 1)
    loc_comb_data[f"{comp_name}_peak_name"] = (
        loc_comb_data["interval"]
        + "_"
        + loc_comb_data[f"{comp_name}_peak_number"].astype(str)
        + "_"
        + loc_comb_data[f"{comp_name}_peak"].astype(str)
    )

    peak_location_df = loc_comb_data[
        loc_comb_data[f"{comp_name}_peak"].values
    ].set_index("location")
    locations = peak_location_df.groupby(f"{comp_name}_peak_name")[comp_name].idxmax()
    peak_groups = peak_location_df.reset_index().groupby(f"{comp_name}_peak_name")
    start = peak_groups["location"].min().values - (step_size / 2)
    end = peak_groups["location"].max().values + (step_size / 2)
    means = peak_groups[comp_name].mean()
    peaks = pd.DataFrame(
        {
            "seqname": seqname,
            "start": start.astype(int),
            "end": end.astype(int),
            "mean": means,
            "summit": locations,
        }
    ).sort_values("start")
    return peaks


def both_tracks_to_intervlas(job):
    loc_comb_data, step_size, seqname = job
    peaks_c1 = track_to_interval(loc_comb_data, "c1", step_size, seqname)
    peaks_c2 = track_to_interval(loc_comb_data, "c2", step_size, seqname)
    return peaks_c1, peaks_c2


def tracks_to_intervals(comb_data, step_size, cores=8, progress=True):
    peaks_list_c1 = list()
    peaks_list_c2 = list()
    jobs = [
        (loc_comb_data, step_size, seqname)
        for seqname, loc_comb_data in comb_data.groupby("seqname")
    ]
    lcores = min(cores, len(jobs))
    logger.debug("Using %d cores for conversion.", lcores)
    with threadpool_limits(limits=1):
        with multiprocessing.Pool(lcores) as pool:
            for peaks_c1, peaks_c2 in tqdm(
                pool.imap(both_tracks_to_intervlas, jobs),
                total=len(jobs),
                disable=not progress,
                desc="sequence",
            ):
                peaks_list_c1.append(peaks_c1)
                peaks_list_c2.append(peaks_c2)
    peaks_c1 = pd.concat(peaks_list_c1, axis=0)
    peaks_c1["length"] = peaks_c1["end"] - peaks_c1["start"]
    peaks_c2 = pd.concat(peaks_list_c2, axis=0)
    peaks_c2["length"] = peaks_c2["end"] - peaks_c2["start"]
    return peaks_c1, peaks_c2


def make_overlap_report(
    comb_data,
    peaks_c1,
    peaks_c2,
    fraction_c1_permissive,
    fraction_c2_permissive,
    progress=True,
):
    sel_comb_data = comb_data[comb_data["c1_peak"] | comb_data["c2_peak"]]
    df_entries = list()
    df_header = [
        "seqname",
        "start",
        "end",
        "length",
        "peak_c1",
        "peak_c2",
        "frac_c1",
        "frac_c2",
        "mean_c1",
        "mean_c2",
        "cosine_distance",
    ]
    for seqname, loc_comb_data in tqdm(
        sel_comb_data.groupby("seqname"), disable=not progress, desc="sequence"
    ):
        loc_peaks_c1 = peaks_c1[peaks_c1["seqname"] == seqname].sort_values("start")
        loc_peaks_c2 = peaks_c2[peaks_c2["seqname"] == seqname].sort_values("start")
        max_pos_p = len(loc_peaks_c1)
        max_pos_k = len(loc_peaks_c2)
        if max_pos_p == 0:
            logger.info(f"No c1 peaks in sequence {seqname}.")
            continue
        if max_pos_k == 0:
            logger.info(f"No c2 peaks in sequence {seqname}.")
            continue
        rows_p = loc_peaks_c1.iterrows()
        name_p, dat_p = next(rows_p)
        rows_k = loc_peaks_c2.iterrows()
        name_k, dat_k = next(rows_k)
        rows_values = loc_comb_data.iterrows()
        _, values = next(rows_values)
        try:
            for _ in range(max_pos_p * max_pos_k):
                if dat_p["start"] > dat_k["end"]:
                    name_k, dat_k = next(rows_k)
                    continue
                if dat_k["start"] < dat_p["end"]:
                    # overlap
                    start = max(dat_k["start"], dat_p["start"])
                    end = min(dat_k["end"], dat_p["end"])
                    while values["location"] < start:
                        # move up towards start of overlap
                        _, values = next(rows_values)
                    values_p = list()
                    values_k = list()
                    while values["location"] < end:
                        # move up towards end of overlap
                        values_p.append(values["c1"])
                        values_k.append(values["c2"])
                        _, values = next(rows_values)
                    if len(values_p) > 1:
                        length = end - start
                        values_p = np.array(values_p)
                        values_k = np.array(values_k)
                        mean_p = np.mean(values_p)
                        mean_k = np.mean(values_k)
                        frac_p = length / (dat_p["end"] - dat_p["start"])
                        frac_k = length / (dat_k["end"] - dat_k["start"])
                        cosine_distance = scipy.spatial.distance.cosine(
                            values_p - mean_p, values_k - mean_k
                        )
                        df_entries.append(
                            (
                                seqname,
                                start,
                                end,
                                length,
                                name_p,
                                name_k,
                                frac_p,
                                frac_k,
                                mean_p,
                                mean_k,
                                cosine_distance,
                            )
                        )
                name_p, dat_p = next(rows_p)
        except StopIteration:
            pass
    overlap_df = pd.DataFrame(df_entries, columns=df_header)
    overlap_df["log_ratio"] = np.log(overlap_df["mean_c1"]) - np.log(
        overlap_df["mean_c2"]
    )
    overlap_df["log_ratio_corrected"] = (
        overlap_df["log_ratio"]
        - np.log(fraction_c1_permissive)
        + np.log(fraction_c2_permissive)
    )
    overlap_df["callability"] = (
        np.abs(overlap_df["log_ratio_corrected"]) * overlap_df["cosine_distance"]
    )
    overlap_df["ambiguous"] = overlap_df["callability"] < 1
    return overlap_df


def merge_overlapping_intervals(interval_df):
    merged_df_list = list()
    for seqname, data in interval_df.groupby("seqname"):
        data = data.sort_values("start")
        last_stop = None
        new_starts = list()
        new_stops = list()
        new_means = list()
        for _, d in data.iterrows():
            start = d["start"]
            stop = d["end"]
            if last_stop is None:  # init
                new_starts.append(start)
                means = [d["mean"]]
                last_stop = stop
                continue
            means.append(d["mean"])
            if start < last_stop:  # combine
                last_stop = max(last_stop, stop)
                continue
            # finish interval
            new_stops.append(last_stop)
            new_means.append(np.mean(means))
            # start new interval
            new_starts.append(start)
            last_stop = stop
        new_stops.append(last_stop)
        new_means.append(np.mean(means))
        merged_df_list.append(
            pd.DataFrame(
                {
                    "seqname": seqname,
                    "start": np.array(new_starts),
                    "end": np.array(new_stops),
                    "mean": np.array(new_means),
                    "name": [f"{seqname}_{i}" for i in range(len(new_starts))],
                }
            )
        )
    return pd.concat(merged_df_list, axis=0)


def fiter_overlaps(peaks_c1, peaks_c2, overlap_df, threshold=0.5):
    fractions_c1 = overlap_df.groupby("peak_c1", sort=False)["frac_c1"].sum()
    fractions_c2 = overlap_df.groupby("peak_c2", sort=False)["frac_c2"].sum()
    bad_idx_c1 = fractions_c1.index[fractions_c1 > threshold]
    overlapping_c1 = peaks_c1.loc[bad_idx_c1, :]
    bad_idx_c2 = fractions_c2.index[fractions_c2 > threshold]
    overlapping_c2 = peaks_c2.loc[bad_idx_c2, :]
    overlap_df = merge_overlapping_intervals(
        pd.concat([overlapping_c1, overlapping_c2])
    )

    peaks_c1 = peaks_c1.loc[peaks_c1.index.difference(bad_idx_c1), :].sort_values(
        ["seqname", "start"]
    )
    peaks_c2 = peaks_c2.loc[peaks_c2.index.difference(bad_idx_c2), :].sort_values(
        ["seqname", "start"]
    )

    return peaks_c1, peaks_c2, overlap_df


def name_peaks(df, prefix):
    name_dat = pd.DataFrame(
        list(df.index.str.split("_")),
        columns=["seqname", "interval", "subinterval", "is_peak"],
        index=df.index,
    )
    df["name"] = (
        str(prefix) + "_"
        + name_dat["seqname"] + "_"
        + name_dat["interval"] + "_"
        + name_dat["subinterval"]
    )
    return df


def inlude_auc(df):
    df["auc"] = df["mean"] * df["length"]
    return df


def bed_to_summit_bed(df):
    summits = df["summit"].astype(int)
    summit_df = pd.DataFrame(
        {
            "seqname": df["seqname"],
            "start": summits,
            "end": summits + 1,
            "name": df["name"],
            "signal": df["mean"],
        }
    )
    return summit_df


def call_peaks(
    comb_data,
    signal_c1_list_global,
    signal_c2_list_global,
    work_coverage=1.0,
    fraction_in_peaks=0.5,
    fraction_overlap=0.5,
    uncorrected=False,
    c1_min_peak_size=100,
    c2_min_peak_size=400,
    span=10,
    cores=16,
    progress=False,
):
    (
        fraction_c1,
        fraction_c2,
        fraction_c1_permissive,
        fraction_c2_permissive,
    ) = target_fractions(comb_data, fraction_in_peaks, uncorrected=uncorrected)
    logger.info("Selecting peak candidates.")
    comb_data = mark_peaks(
        comb_data,
        signal_c1_list_global,
        signal_c2_list_global,
        fraction_c1,
        fraction_c2,
        work_coverage,
    )
    logger.info("Converting peak signal to intervals.")
    peaks_c1, peaks_c2 = tracks_to_intervals(comb_data, span, cores, progress=progress)
    peaks_c1 = peaks_c1[peaks_c1["length"] > c1_min_peak_size]
    peaks_c2 = peaks_c2[peaks_c2["length"] > c2_min_peak_size]

    logger.info("Analyzing overlaps.")
    overlap_df = make_overlap_report(
        comb_data,
        peaks_c1,
        peaks_c2,
        fraction_c1_permissive,
        fraction_c2_permissive,
        progress=progress,
    )

    logger.info("Filtering overlaps.")
    peaks_c1, peaks_c2, overlaps = fiter_overlaps(
        peaks_c1, peaks_c2, overlap_df, threshold=fraction_overlap
    )
    logger.info("Peaks for c1: %d", len(peaks_c1))
    logger.info("Peaks for c2: %d", len(peaks_c2))
    return peaks_c1, peaks_c2, overlaps


def write_peaks(bed_c1, bed_c2, overlaps, out_path):

    out_base = os.path.join(out_path, "peaks_")

    write_bed(
        bed_c1[["seqname", "start", "end", "name", "auc"]], out_base + "deconv_c1.bed"
    )
    write_bed(
        bed_c2[["seqname", "start", "end", "name", "auc"]], out_base + "deconv_c2.bed"
    )

    write_bed(bed_to_summit_bed(bed_c1), out_base + "deconv_c1.summits.bed")
    write_bed(bed_to_summit_bed(bed_c2), out_base + "deconv_c2.summits.bed")

    write_bed(
        overlaps[["seqname", "start", "end", "name", "mean"]], out_base + "overlap.bed",
    )
    return


def main():
    args = parser.parse_args()
    setup_logging(args.logLevel, args.logfile)
    logger.debug("Loglevel is on DEBUG.")
    cores = detect_cores(args.cores)
    logger.info("Limiting computation to %i cores.", cores)
    threadpool_limits(limits=int(cores))
    os.environ["NUMEXPR_MAX_THREADS"] = str(cores)

    logger.info("Reading jobdata.")
    workdata = read_job_data(args.jobdata)
    work_coverage = estimated_deconvolved_fraction(workdata)
    logger.info(f"At most {work_coverage:.2%} of the genome is deconvolved.")
    logger.info("Reading deconvolution results.")
    try:
        map_results = read_results(
            args.jobdata, workdata, progress=not args.no_progress, error=not args.force
        )
    except MissingData:
        raise MissingData(
            "Results are missing. Complete missing work chunks or pass --force to "
            "call peaks regardless."
        )

    if not args.no_check:
        check_length_distribution_flip(workdata, map_results)

    if args.only_check:
        return

    logger.info("Processing deconvolved tracks.")
    comb_data, signal_c1_list_global, signal_c2_list_global = interpolate(
        workdata,
        map_results,
        args.c1_smooth,
        args.c2_smooth,
        args.span,
        cores=cores,
        progress=not args.no_progress,
    )

    logger.info("Calling peaks.")
    peaks_c1, peaks_c2, overlaps = call_peaks(
        comb_data,
        signal_c1_list_global,
        signal_c2_list_global,
        work_coverage=work_coverage,
        fraction_in_peaks=args.fraction_in_peaks,
        fraction_overlap=args.fraction_overlap,
        uncorrected=args.uncorrected,
        c1_min_peak_size=args.c1_min_peak_size,
        c2_min_peak_size=args.c2_min_peak_size,
        span=args.span,
        cores=cores,
        progress=not args.no_progress,
    )

    bed_c1 = inlude_auc(name_peaks(peaks_c1, "c1"))
    bed_c2 = inlude_auc(name_peaks(peaks_c2, "c2"))

    if args.out is None:
        out_path, old_file = os.path.split(args.jobdata)
    else:
        out_path = args.out
        try:
            os.mkdir(out_path)
        except FileExistsError:
            pass

    logger.info("Writing results to %s/...", out_path)
    write_peaks(bed_c1, bed_c2, overlaps, out_path)
    logger.info("Finished sucesfully.")


if __name__ == "__main__":
    main()
